import os
from dataclasses import asdict
from typing import Any, Dict, Optional, Type
from urllib.parse import quote

from platypush.plugins import action
from platypush.plugins.media import PlayerState, MediaPlugin
from platypush.message.event.media import (
    MediaEvent,
    MediaPlayEvent,
    MediaPlayRequestEvent,
    MediaPauseEvent,
    MediaStopEvent,
    NewPlayingMediaEvent,
    MediaSeekEvent,
    MediaResumeEvent,
)


class MediaMpvPlugin(MediaPlugin):
    """
    Plugin to control MPV instances.
    """

    _default_mpv_args = {
        'ytdl': True,
        'start_event_thread': True,
    }

    def __init__(
        self, args: Optional[Dict[str, Any]] = None, fullscreen: bool = False, **kwargs
    ):
        """
        :param args: Default arguments that will be passed to the mpv executable
            as a key-value dict (names without the `--` prefix). See `man mpv`
            for available options.
        :param fullscreen: Set to True if you want media files to be opened in
            fullscreen by default (can be overridden by `.play()`) (default: False)
        """

        super().__init__(**kwargs)

        self.args = {**self._default_mpv_args}
        if args:
            self.args.update(args)
        if fullscreen:
            self.args['fs'] = True

        self._player = None
        self._latest_state = PlayerState.STOP

    def _init_mpv(self, args=None):
        import mpv

        mpv_args = self.args.copy()
        if args:
            mpv_args.update(args)

        for k, v in self._env.items():
            os.environ[k] = v

        self._player = mpv.MPV(**mpv_args)
        self._player._event_callbacks += [self._event_callback()]

    def _post_event(self, evt_type: Type[MediaEvent], **evt):
        self._bus.post(
            evt_type(
                player='local',
                plugin='media.mpv',
                resource=evt.pop('resource', self._resource),
                title=self._filename,
                **evt,
            )
        )

    @property
    def _cur_player(self):
        if self._player and not self._player.core_shutdown:
            return self._player

        return None

    @property
    def _state(self):
        player = self._cur_player
        if not player:
            return PlayerState.STOP

        return PlayerState.PAUSE if player.pause else PlayerState.PLAY

    @property
    def _resource(self):
        if not self._cur_player:
            return None

        cur_resource = self._cur_player.stream_path
        if not cur_resource:
            return None

        return quote(
            ('file://' if os.path.isfile(cur_resource) else '') + str(cur_resource)
        )

    @property
    def _filename(self):
        if not self._cur_player:
            return None

        return self._cur_player.filename

    def _event_callback(self):
        def callback(event):
            from mpv import MpvEvent

            self.logger.info('Received mpv event: %s', event)

            if isinstance(event, MpvEvent):
                event = event.as_dict()
            if not isinstance(event, dict):
                return

            evt_type = event.get('event', b'').decode()
            if not evt_type:
                return

            if evt_type == 'start-file':
                self._post_event(NewPlayingMediaEvent)
            elif evt_type == 'playback-restart':
                self._post_event(MediaPlayEvent)
            elif evt_type in ('shutdown', 'idle', 'end-file'):
                if self._state != PlayerState.PLAY:
                    self._post_event(MediaStopEvent)

                if evt_type == 'shutdown' and self._player:
                    self._player = None
            elif evt_type == 'seek' and self._cur_player:
                self._post_event(
                    MediaSeekEvent, position=self._cur_player.playback_time
                )

            self._latest_state = self._state

        return callback

    @action
    def execute(self, cmd, **args):
        """
        Execute a raw mpv command.
        """
        if not self._cur_player:
            return None

        return self._cur_player.command(cmd, *args)

    @action
    def play(
        self,
        resource: str,
        *_,
        subtitles: Optional[str] = None,
        fullscreen: Optional[bool] = None,
        **args,
    ):
        """
        Play a resource.

        :param resource: Resource to play - can be a local file or a remote URL
        :param subtitles: Path to optional subtitle file
        :param args: Extra runtime arguments that will be passed to the
            mpv executable as a key-value dict (keys without `--` prefix)
        """

        self._post_event(MediaPlayRequestEvent, resource=resource)
        if fullscreen is not None:
            args['fs'] = fullscreen

        self._init_mpv(args)

        resource = self._get_resource(resource)
        if resource.startswith('file://'):
            resource = resource[7:]

        assert self._cur_player, 'The player is not ready'
        self._cur_player.play(resource)
        if self.volume:
            self.set_volume(volume=self.volume)
        if subtitles:
            self.add_subtitles(subtitles)

        return self.status()

    @action
    def pause(self, *_, **__):
        """Toggle the paused state"""
        if not self._cur_player:
            return None

        self._cur_player.pause = not self._cur_player.pause
        return self.status()

    @action
    def quit(self, *_, **__):
        """Stop and quit the player"""
        player = self._cur_player
        if not player:
            return None

        player.stop()
        player.quit(code=0)
        player.wait_for_shutdown(timeout=10)
        player.terminate()
        self._player = None
        return self.status()

    @action
    def stop(self, *_, **__):
        """Stop and quit the player"""
        return self.quit()

    def _set_vol(self, *_, step=10.0, **__):
        if not self._cur_player:
            return None

        return self.set_volume(float(self._cur_player.volume or 0) - step)

    @action
    def voldown(self, *_, step: float = 10.0, **__):
        """Volume down by (default: 10)%"""
        if not self._cur_player:
            return None

        return self.set_volume(float(self._cur_player.volume or 0) - step)

    @action
    def volup(self, step: float = 10.0, **_):
        """Volume up by (default: 10)%"""
        if not self._cur_player:
            return None

        return self.set_volume(float(self._cur_player.volume or 0) + step)

    @action
    def set_volume(self, volume):
        """
        Set the volume

        :param volume: Volume value between 0 and 100
        :type volume: float
        """
        if not self._cur_player:
            return None

        max_vol = (
            self._cur_player.volume_max
            if self._cur_player.volume_max is not None
            else 100
        )
        volume = max(0, min([max_vol, volume]))
        self._cur_player.volume = volume
        return self.status()

    @action
    def seek(self, position: float, **_):
        """
        Seek backward/forward by the specified number of seconds

        :param position: Number of seconds relative to the current cursor
        """
        if not self._cur_player:
            return None

        assert self._cur_player.seekable, 'The resource is not seekable'
        self._cur_player.time_pos = min(
            float(self._cur_player.time_pos or 0)
            + float(self._cur_player.time_remaining or 0),
            max(0, position),
        )
        return self.status()

    @action
    def back(self, offset=30.0, **_):
        """Back by (default: 30) seconds"""
        if not self._cur_player:
            return None

        assert self._cur_player.seekable, 'The resource is not seekable'
        cur_pos = float(self._cur_player.time_pos or 0)
        return self.seek(cur_pos - offset)

    @action
    def forward(self, offset=30.0, **_):
        """Forward by (default: 30) seconds"""
        if not self._cur_player:
            return None

        assert self._cur_player.seekable, 'The resource is not seekable'
        cur_pos = float(self._cur_player.time_pos or 0)
        return self.seek(cur_pos + offset)

    @action
    def next(self, **_):
        """Play the next item in the queue"""
        if not self._cur_player:
            return None

        self._cur_player.playlist_next()
        return self.status()

    @action
    def prev(self, **_):
        """Play the previous item in the queue"""
        if not self._cur_player:
            return None

        self._cur_player.playlist_prev()
        return self.status()

    @action
    def toggle_subtitles(self, *_, **__):
        """Toggle the subtitles visibility"""
        return self.toggle_property('sub_visibility')

    @action
    def add_subtitles(self, filename):
        """Add a subtitles file"""
        if not self._cur_player:
            return None

        return self._cur_player.sub_add(filename)

    @action
    def toggle_fullscreen(self):
        """Toggle the fullscreen mode"""
        return self.toggle_property('fullscreen')

    @action
    def toggle_property(self, property):
        """
        Toggle or sets the value of an mpv property (e.g. fullscreen,
        sub_visibility etc.). See ``man mpv`` for a full list of properties

        :param property: Property to toggle
        """
        if not self._player:
            return None

        if not hasattr(self._player, property):
            self.logger.warning('No such mpv property: {}'.format(property))

        value = not getattr(self._player, property)
        setattr(self._player, property, value)
        return {property: value}

    @action
    def get_property(self, property):
        """
        Get a player property (e.g. pause, fullscreen etc.). See
        ``man mpv`` for a full list of the available properties
        """
        if not self._player:
            return None
        return getattr(self._player, property)

    @action
    def set_property(self, **props):
        """
        Set the value of an mpv property (e.g. fullscreen, sub_visibility
        etc.). See ``man mpv`` for a full list of properties

        :param props: Key-value args for the properties to set
        :type props: dict
        """
        if not self._player:
            return None

        for k, v in props.items():
            setattr(self._player, k, v)
        return props

    @action
    def set_subtitles(self, filename, *_, **__):
        """Sets media subtitles from filename"""
        return self.set_property(subfile=filename, sub_visibility=True)

    @action
    def remove_subtitles(self, sub_id=None, **_):
        """Removes (hides) the subtitles"""
        if not self._player:
            return None
        if sub_id:
            return self._player.sub_remove(sub_id)
        self._player.sub_visibility = False

    @action
    def is_playing(self, **_):
        """
        :returns: True if it's playing, False otherwise
        """
        if not self._player:
            return False
        return not self._player.pause

    @action
    def load(self, resource, **args):
        """
        Load/queue a resource/video to the player
        """
        if not self._player:
            return self.play(resource, **args)
        return self._player.loadfile(resource, mode='append-play')

    @action
    def mute(self, **_):
        """Toggle mute state"""
        if not self._player:
            return None
        mute = not self._player.mute
        self._player.mute = mute
        return {'muted': mute}

    @action
    def set_position(self, position: float, **_):
        """
        Seek backward/forward to the specified absolute position (same as ``seek``)
        """
        return self.seek(position)

    @action
    def status(self, **_):
        """
        Get the current player state.

        :returns: A dictionary containing the current state.

        Example:

            .. code-block:: javascript

                {
                    "audio_channels": 2,
                    "audio_codec": "mp3",
                    "delay": 0,
                    "duration": 300.0,
                    "file_size": 123456,
                    "filename": "filename or stream URL",
                    "fullscreen": false,
                    "mute": false,
                    "name": "mpv",
                    "pause": false,
                    "percent_pos": 10.0,
                    "position": 30.0,
                    "seekable": true,
                    "state": "play",  // or "stop" or "pause"
                    "title": "filename or stream URL",
                    "url": "file:///path/to/file.mp3",
                    "video_codec": "h264",
                    "video_format": "avc1",
                    "volume": 50.0,
                    "volume_max": 100.0,
                    "width": 1280
                }

        """
        if not self._cur_player:
            return {'state': PlayerState.STOP.value}

        status = {
            'audio_channels': getattr(self._player, 'audio_channels', None),
            'audio_codec': getattr(self._player, 'audio_codec_name', None),
            'delay': getattr(self._player, 'delay', None),
            'duration': (
                (getattr(self._player, 'playback_time', 0) or 0)
                + getattr(self._player, 'playtime_remaining', 0)
                if getattr(self._player, 'playtime_remaining', None)
                else None
            ),
            'filename': getattr(self._player, 'filename', None),
            'file_size': getattr(self._player, 'file_size', None),
            'fullscreen': getattr(self._player, 'fs', None),
            'mute': getattr(self._player, 'mute', None),
            'name': getattr(self._player, 'name', None),
            'pause': getattr(self._player, 'pause', None),
            'percent_pos': getattr(self._player, 'percent_pos', None),
            'position': getattr(self._player, 'playback_time', None),
            'seekable': getattr(self._player, 'seekable', None),
            'state': self._state.value,
            'title': getattr(self._player, 'media_title', None)
            or getattr(self._player, 'filename', None),
            'url': self._resource,
            'video_codec': getattr(self._player, 'video_codec', None),
            'video_format': getattr(self._player, 'video_format', None),
            'volume': getattr(self._player, 'volume', None),
            'volume_max': getattr(self._player, 'volume_max', None),
            'width': getattr(self._player, 'width', None),
        }

        if self._latest_resource:
            status.update(
                {
                    k: v
                    for k, v in asdict(self._latest_resource).items()
                    if v is not None
                }
            )

        if self._state != self._latest_state:
            if not self._cur_player:
                self._post_event(MediaStopEvent)
            else:
                self._post_event(
                    MediaPauseEvent
                    if self._state == PlayerState.PAUSE
                    else MediaResumeEvent
                )

        self._latest_state = self._state
        return status

    def _get_resource(self, resource):
        if self._is_youtube_resource(resource):
            return resource  # mpv can handle YouTube streaming natively

        return super()._get_resource(resource)


# vim:sw=4:ts=4:et:



import logging
from typing import Iterable, List, Optional

from sqlalchemy.orm import Session
from sqlalchemy.orm.exc import ObjectDeletedError

from platypush.entities._base import Entity, EntityMapping

from .helpers import get_parent

logger = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class EntitiesMerger:
    """
    A stateless functor in charge of detecting and merging entities that
    already exist on the database before flushing the session.
    """

    def __call__(
        self,
        session: Session,
        entities: Iterable[Entity],
        existing_entities: Optional[EntityMapping] = None,
    ) -> List[Entity]:
        """
        Merge a set of entities with their existing representations and update
        the parent/child relationships and return a list containing
        ``[*updated_entities, *new_entities]``.
        """
        existing_entities = existing_entities or {}
        new_entities: EntityMapping = {}

        self._merge(
            session,
            entities,
            new_entities=new_entities,
            existing_entities=existing_entities,
        )

        return list({**existing_entities, **new_entities}.values())

    def _merge(
        self,
        session: Session,
        entities: Iterable[Entity],
        new_entities: EntityMapping,
        existing_entities: EntityMapping,
    ) -> List[Entity]:
        """
        (Recursive) inner implementation of the entity merge logic.
        """
        processed_entities = []

        # Retrieve existing records and merge them
        for entity in entities:
            key = entity.entity_key
            existing_entity = existing_entities.get(key, new_entities.get(key))

            # Synchronize the parent(s)
            entity = self._sync_parent(session, entity, new_entities, existing_entities)

            if existing_entity:
                # Merge the columns with those of the existing entity
                existing_entity = self._merge_columns(entity, existing_entity)
                # Merge the children
                self._append_children(
                    existing_entity,
                    *self._merge(
                        session,
                        entity.children,
                        new_entities,
                        existing_entities,
                    )
                )

                # Use the existing entity now that it's been merged
                entity = existing_entity
            else:
                # Add it to the map of new entities if the entity doesn't exist on the db
                new_entities[key] = entity

            processed_entities.append(entity)

        return processed_entities

    @classmethod
    def _sync_parent(
        cls,
        session: Session,
        entity: Entity,
        new_entities: EntityMapping,
        existing_entities: EntityMapping,
    ) -> Entity:
        """
        Recursively refresh the parent of an entity all the way up in the
        hierarchy, to make sure that all the parent/child relations are
        appropriately rewired and that all the relevant objects are added to
        this session.
        """
        parent = get_parent(session, entity)
        if not parent:
            # No parent -> we can terminate the recursive climbing
            return entity

        # Check if an entity with the same key as the reported parent already
        # exists in the cached entities
        existing_parent = existing_entities.get(
            parent.entity_key, new_entities.get(parent.entity_key)
        )

        if not existing_parent:
            # No existing parent -> we need to flush the one reported by this
            # entity
            return entity

        # Check if the existing parent already has a child with the same key as
        # this entity
        existing_entity = next(
            iter(
                child
                for child in existing_parent.children
                if child.entity_key == entity.entity_key
            ),
            None,
        )

        if not existing_entity:
            # If this entity isn't currently a member of the existing parent,
            # temporarily reset the parent of the current entity, so we won't
            # carry stale objects around. We will soon rewire it to the
            # existing parent.
            entity.parent = None
        else:
            # Otherwise, merge the columns of the existing entity with those of
            # the new entity and use the existing entity
            entity = cls._merge_columns(entity, existing_entity, include_children=True)

        # Refresh the existing collection of children with the new/updated
        # entity
        cls._append_children(existing_parent, entity)

        # Recursively call this function to synchronize any parent entities up
        # in the taxonomy
        cls._sync_parent(session, existing_parent, new_entities, existing_entities)
        return entity

    @staticmethod
    def _append_children(entity: Entity, *children: Entity):
        """
        Update the list of children of a given entity with the given list of
        entities.

        Note that, in case of ``entity_key`` conflict (the key of a new entity
        already exists in the entity's children), the most recent version will
        be used, so any column merge logic needs to happen before this method
        is called.
        """
        entity.children = list(
            {
                **{e.entity_key: e for e in entity.children},
                **{e.entity_key: e for e in children},
            }.values()
        )

        for child in children:
            child.parent = entity
            if entity.id:
                child.parent_id = entity.id

    @classmethod
    def _merge_columns(
        cls, entity: Entity, existing_entity: Entity, include_children: bool = False
    ) -> Entity:
        """
        Merge two versions of an entity column by column.
        """
        columns = [col.key for col in entity.get_columns()]
        for col in columns:
            if col == 'meta':
                existing_entity.meta = {  # type: ignore
                    **(existing_entity.meta or {}),  # type: ignore
                    **(entity.meta or {}),  # type: ignore
                }
            elif col not in ('id', 'created_at'):
                try:
                    setattr(existing_entity, col, getattr(entity, col))
                except ObjectDeletedError as e:
                    logger.warning(
                        'Could not set %s on entity <%s>: %s',
                        col,
                        existing_entity.entity_key,
                        e,
                    )

        # Recursive call to merge the columns of the children too
        if include_children:
            existing_children = {e.entity_key: e for e in existing_entity.children}
            new_children = {e.entity_key: e for e in entity.children}
            updated_children = {}

            for key, child in new_children.items():
                existing_child = existing_children.get(key)
                updated_children[key] = (
                    cls._merge_columns(child, existing_child, include_children=True)
                    if existing_child
                    else child
                )

            cls._append_children(existing_entity, *updated_children.values())

        return existing_entity



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from threading import Event
from typing import Any, Callable, Optional, Type


class InviteTarget(Enum):
    """
    Tells whether the target of an invite is a user or a room.
    """

    USER = 1
    ROOM = 2


@dataclass
class Invite(ABC):
    """
    A class that models the parameters of an invite to a conversation.
    """

    accepted: Optional[bool] = None
    responded: Event = field(default_factory=Event)
    on_accepted: Callable[[], Any] = field(default_factory=lambda: lambda: None)
    on_rejected: Callable[[], Any] = field(default_factory=lambda: lambda: None)

    @property
    @abstractmethod
    def target(self) -> InviteTarget:
        raise NotImplementedError

    def accept(self):
        self.accepted = True
        self.responded.set()
        self.on_accepted()

    def reject(self):
        self.accepted = False
        self.responded.set()
        self.on_rejected()

    def wait_response(self, timeout: Optional[float] = None) -> bool:
        return self.responded.wait(timeout)

    @classmethod
    def by_target(cls, target: InviteTarget) -> Type["Invite"]:
        return {
            InviteTarget.ROOM: RoomInvite,
            InviteTarget.USER: UserInvite,
        }[target]


@dataclass
class RoomInvite(Invite):
    """
    Models an invite to a room.
    """

    @property
    def target(self) -> InviteTarget:
        return InviteTarget.ROOM


@dataclass
class UserInvite(Invite):
    """
    Models an invite to a user's contacts list.
    """

    @property
    def target(self) -> InviteTarget:
        return InviteTarget.USER



from marshmallow import EXCLUDE, fields
from marshmallow.schema import Schema


class TrelloLabelSchema(Schema):
    """
    Trello label schema.
    """

    id = fields.String(
        required=True,
        metadata={
            "description": "The label's unique identifier.",
            "example": "5d62808da5d6a95a3a3e4f2f",
        },
    )

    name = fields.String(
        required=True,
        metadata={
            "description": "The label's name.",
            "example": "My Label",
        },
    )

    color = fields.String(
        metadata={
            "description": "The label's color.",
            "example": "green",
        },
    )


class TrelloUserSchema(Schema):
    """
    Trello user schema.
    """

    id = fields.String(
        required=True,
        metadata={
            "description": "The user's unique identifier.",
            "example": "5d62808da5d6a95a3a3e4f2f",
        },
    )

    username = fields.String(
        required=True,
        metadata={
            "description": "The user's username.",
            "example": "myusername",
        },
    )

    fullname = fields.String(
        metadata={
            "description": "The user's full name.",
            "example": "My Full Name",
        },
    )

    initials = fields.String(
        metadata={
            "description": "The user's initials.",
            "example": "MFN",
        },
    )

    avatar_url = fields.Url(
        metadata={
            "description": "The user's avatar URL.",
            "example": "https://trello-avatars.s3.amazonaws.com/5d62808da5d6a95a3a3e4f2f/50.png",
        },
    )


class TrelloMemberSchema(Schema):
    """
    Trello member schema.
    """

    id = fields.String(
        required=True,
        metadata={
            "description": "The user's unique identifier.",
            "example": "5d62808da5d6a95a3a3e4f2f",
        },
    )

    username = fields.String(
        required=True,
        metadata={
            "description": "The user's username.",
            "example": "myusername",
        },
    )

    fullname = fields.String(
        metadata={
            "description": "The user's full name.",
            "example": "My Full Name",
        },
    )

    initials = fields.String(
        metadata={
            "description": "The user's initials.",
            "example": "MFN",
        },
    )

    bio = fields.String(
        metadata={
            "description": "The user's bio.",
            "example": "My bio.",
        },
    )

    member_type = fields.String(
        metadata={
            "description": "The user's type.",
            "example": "admin",
        },
    )


class TrelloCommentSchema(Schema):
    """
    Trello comment schema.
    """

    id = fields.String(
        required=True,
        metadata={
            "description": "The comment's unique identifier.",
            "example": "5d62808da5d6a95a3a3e4f2f",
        },
    )

    text = fields.String(
        required=True,
        metadata={
            "description": "The comment's text.",
            "example": "My comment's text.",
        },
    )

    type = fields.String(
        required=True,
        metadata={
            "description": "The comment's type.",
            "example": "commentCard",
        },
    )

    creator = fields.Nested(TrelloUserSchema)
    date = fields.DateTime(
        metadata={
            "description": "The comment's date.",
            "example": "2019-08-25T15:32:13.000Z",
        },
    )


class TrelloListSchema(Schema):
    """
    Trello list schema.
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta class.
        """

        unknown = EXCLUDE

    id = fields.String(
        required=True,
        metadata={
            "description": "The list's unique identifier.",
            "example": "5d62808da5d6a95a3a3e4f2f",
        },
    )

    name = fields.String(
        required=True,
        metadata={
            "description": "The list's name.",
            "example": "My List",
        },
    )

    closed = fields.Boolean(
        missing=False,
        metadata={
            "description": "Whether the list is closed.",
            "example": False,
        },
    )

    subscribed = fields.Boolean(
        missing=False,
        metadata={
            "description": "Whether the list is subscribed.",
            "example": False,
        },
    )


class TrelloBoardSchema(Schema):
    """
    Trello board schema.
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta class.
        """

        unknown = EXCLUDE

    id = fields.String(
        required=True,
        metadata={
            "description": "The board's unique identifier.",
            "example": "5d62808da5d6a95a3a3e4f2f",
        },
    )

    name = fields.String(
        required=True,
        metadata={
            "description": "The board's name.",
            "example": "My Board",
        },
    )

    url = fields.Url(
        required=True,
        metadata={
            "description": "The board's URL.",
            "example": "https://trello.com/b/5d62808da5d6a95a3a3e4f2f/my-board",
        },
    )

    closed = fields.Boolean(
        missing=False,
        metadata={
            "description": "Whether the board is closed.",
            "example": False,
        },
    )

    lists = fields.Nested(
        TrelloListSchema,
        many=True,
        metadata={
            "description": "The board's lists.",
        },
    )

    date_last_activity = fields.DateTime(
        metadata={
            "description": "The board's last activity date.",
            "example": "2019-08-25T15:52:45.000Z",
        },
    )


class TrelloPreviewSchema(Schema):
    """
    Trello attachment preview schema.
    """

    id = fields.String(
        required=True,
        metadata={
            "description": "The preview's unique identifier.",
            "example": "5d62808da5d6a95a3a3e4f2f",
        },
    )

    url = fields.Url(
        required=True,
        metadata={
            "description": "The preview's URL.",
            "example": "https://trello.com/c/5d62808da5d6a95a3a3e4f2f/my-attachment-txt.jpg",
        },
    )

    scaled = fields.Boolean(
        metadata={
            "description": "Whether the preview is scaled.",
            "example": True,
        },
    )

    size = fields.Integer(
        metadata={
            "description": "The preview's size, in bytes.",
            "example": 10000,
        },
    )

    width = fields.Integer(
        metadata={
            "description": "The preview's width, in pixels.",
            "example": 100,
        },
    )

    height = fields.Integer(
        metadata={
            "description": "The preview's height, in pixels.",
            "example": 100,
        },
    )


class TrelloChecklistItemSchema(Schema):
    """
    Trello checklist item schema.
    """

    id = fields.String(
        required=True,
        metadata={
            "description": "The checklist item's unique identifier.",
            "example": "5d62808da5d6a95a3a3e4f2f",
        },
    )

    name = fields.String(
        required=True,
        metadata={
            "description": "The checklist item's name.",
            "example": "My Checklist Item",
        },
    )

    checked = fields.Boolean(
        metadata={
            "description": "Whether the checklist item is checked.",
            "example": True,
        },
    )


class TrelloChecklistSchema(Schema):
    """
    Trello checklist schema.
    """

    id = fields.String(
        required=True,
        metadata={
            "description": "The checklist's unique identifier.",
            "example": "5d62808da5d6a95a3a3e4f2f",
        },
    )

    name = fields.String(
        metadata={
            "description": "The checklist's name.",
            "example": "My Checklist",
        },
    )

    items = fields.Nested(TrelloChecklistItemSchema, many=True)


class TrelloAttachmentSchema(Schema):
    """
    Trello attachment schema.
    """

    id = fields.String(
        required=True,
        metadata={
            "description": "The attachment's unique identifier.",
            "example": "5d62808da5d6a95a3a3e4f2f",
        },
    )

    name = fields.String(
        required=True,
        metadata={
            "description": "The attachment's name.",
            "example": "My Attachment.txt",
        },
    )

    url = fields.Url(
        required=True,
        metadata={
            "description": "The attachment's URL.",
            "example": "https://trello.com/c/5d62808da5d6a95a3a3e4f2f/my-attachment.txt",
        },
    )

    size = fields.Integer(
        metadata={
            "description": "The attachment's size, in bytes.",
            "example": 1024,
        },
    )

    date = fields.DateTime(
        metadata={
            "description": "The attachment's date.",
            "example": "2019-08-25T15:32:13.000Z",
        },
    )

    edge_color = fields.String(
        metadata={
            "description": "The attachment's edge color.",
            "example": "#000000",
        },
    )

    member_id = fields.String(
        metadata={
            "description": "The ID of the member who created the attachment.",
            "example": "5d62808da5d6a95a3a3e4f2f",
        },
    )

    is_upload = fields.Boolean(
        metadata={
            "description": "Whether the attachment is an upload.",
            "example": True,
        },
    )

    mime_type = fields.String(
        metadata={
            "description": "The attachment's MIME type.",
            "example": "text/plain",
        },
    )

    previews = fields.Nested(TrelloPreviewSchema, many=True)


class TrelloCardSchema(Schema):
    """
    Trello card schema.
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta class.
        """

        unknown = EXCLUDE

    id = fields.String(
        required=True,
        metadata={
            "description": "The card's unique identifier.",
            "example": "5d62808da5d6a95a3a3e4f2f",
        },
    )

    name = fields.String(
        required=True,
        metadata={
            "description": "The card's name.",
            "example": "My Card",
        },
    )

    description = fields.String(
        metadata={
            "description": "The card's description.",
            "example": "My card's description.",
        },
    )

    url = fields.Url(
        required=True,
        metadata={
            "description": "The card's URL.",
            "example": "https://trello.com/c/5d62808da5d6a95a3a3e4f2f/my-card",
        },
    )

    due_date = fields.DateTime(
        metadata={
            "description": "The card's due date.",
            "example": "2019-08-25T15:52:45.000Z",
        },
    )

    latest_card_move_date = fields.DateTime(
        metadata={
            "description": "The card's latest move date.",
            "example": "2019-08-25T15:52:45.000Z",
        },
    )

    date_last_activity = fields.DateTime(
        metadata={
            "description": "The card's last activity date.",
            "example": "2019-08-25T15:52:45.000Z",
        },
    )

    closed = fields.Boolean(
        missing=False,
        metadata={
            "description": "Whether the card is closed.",
            "example": False,
        },
    )

    is_due_complete = fields.Boolean(
        metadata={
            "description": "Whether the card is due complete.",
            "example": False,
        },
    )

    board = fields.Nested(TrelloBoardSchema)
    list = fields.Nested(TrelloListSchema)
    comments = fields.Nested(TrelloCommentSchema, many=True)
    labels = fields.Nested(TrelloLabelSchema, many=True)
    attachments = fields.Nested(TrelloAttachmentSchema, many=True)
    checklists = fields.Nested(TrelloChecklistSchema, many=True)



import numpy as np


def convert_nd_array(wave, dtype):  # pylint: disable=too-many-return-statements
    """
    Given a wave as a series of floating point numbers, convert them to the
    appropriate data type.
    """
    t = np.dtype(dtype)
    if t in {np.float16, np.float32, np.float64}:
        return wave.astype(t)
    if t == np.int8:
        return (wave * 2**8).astype(t)
    if t == np.uint8:
        return ((wave + 1) * 2**8).astype(t)
    if t == np.int16:
        return (wave * 2**15).astype(t)
    if t == np.uint16:
        return ((wave + 1) * 2**16).astype(t)
    if t == np.int32:
        return (wave * 2**31).astype(t)
    if t == np.uint32:
        return ((wave + 1) * 2**32).astype(t)

    raise AssertionError(f'Unsupported dtype: {dtype}')



import queue
import threading
from urllib.parse import quote_plus
from typing import Iterable, List, Optional, Union

import requests

from ._base import TorrentSearchProvider
from ._model import TorrentSearchResult


class PopcornTimeSearchProvider(TorrentSearchProvider):
    """
    Torrent search provider that uses the PopcornTime API to search for movies
    and TV shows.
    """

    _http_timeout = 20

    def __init__(
        self,
        api_url: str,
        **kwargs,
    ):
        """
        :param api_url: PopcornTime API base URL.
        """
        super().__init__(**kwargs)
        self.api_url = api_url
        self.torrent_base_urls = {
            'movies': f'{api_url}/movie/{{}}',
            'tv': f'{api_url}/show/{{}}',
        }

    @classmethod
    def provider_name(cls) -> str:
        return 'popcorntime'

    def _search(
        self,
        query: str,
        *args,
        category: Optional[Union[str, Iterable[str]]] = None,
        language: Optional[str] = None,
        **kwargs,
    ) -> List[TorrentSearchResult]:
        """
        Perform a search of video torrents using the PopcornTime API.

        :param query: Query string, video name or partial name
        :param category: Category to search. Supported types: `movies`, `tv`.
            Default: None (search all categories)
        :param language: Language code for the results - example: "en" (default: None, no filter)
        """

        results = []
        if isinstance(category, str):
            category = [category]

        def worker(cat):
            if cat not in self.categories:
                raise RuntimeError(
                    f'Unsupported category {cat}. Supported categories: '
                    f'{list(self.categories.keys())}'
                )

            self.logger.info('Searching %s torrents for "%s"', cat, query)
            results.extend(
                self.categories[cat](self, query, *args, language=language, **kwargs)
            )

        workers = [
            threading.Thread(target=worker, kwargs={'cat': category})
            for category in (category or self.categories.keys())
        ]

        for wrk in workers:
            wrk.start()
        for wrk in workers:
            wrk.join()

        return results

    def _imdb_query(self, query: str, category: str):
        if not query:
            return []

        if category == 'movies':
            imdb_category = 'movie'
        elif category == 'tv':
            imdb_category = 'tvSeries'
        else:
            raise RuntimeError(f'Unsupported category: {category}')

        imdb_url = f'https://v3.sg.media-imdb.com/suggestion/x/{quote_plus(query)}.json?includeVideos=1'
        response = requests.get(imdb_url, timeout=self._http_timeout)
        response.raise_for_status()
        response = response.json()
        assert not response.get('errorMessage'), response['errorMessage']
        return [
            item for item in response.get('d', []) if item.get('qid') == imdb_category
        ]

    def _torrent_search_worker(self, imdb_id: str, category: str, q: queue.Queue):
        base_url = self.torrent_base_urls.get(category)
        assert base_url, f'No such category: {category}'
        try:
            self.logger.debug('Searching torrents for %s', imdb_id)
            response = requests.get(
                base_url.format(imdb_id), timeout=self._http_timeout
            )

            self.logger.debug('Got response for %s: %s', imdb_id, response.text)
            result = response.json()
            q.put(
                result if result.get('code') != 404 and result.get('imdb_id') else None
            )
        except Exception as e:
            q.put(e)

    def _search_torrents(self, query, category):
        imdb_results = self._imdb_query(query, category)
        result_queues = [queue.Queue()] * len(imdb_results)
        workers = [
            threading.Thread(
                target=self._torrent_search_worker,
                kwargs={
                    'imdb_id': imdb_results[i]['id'],
                    'category': category,
                    'q': result_queues[i],
                },
            )
            for i in range(len(imdb_results))
        ]

        results = []
        errors = []

        for worker in workers:
            worker.start()
        for q in result_queues:
            res_ = q.get()
            if isinstance(res_, Exception):
                errors.append(res_)
            else:
                results.append(res_)
        for worker in workers:
            worker.join()

        if errors:
            self.logger.warning('Torrent search errors: %s', [str(e) for e in errors])

        return results

    @classmethod
    def _results_to_movies_response(
        cls, results: List[dict], language: Optional[str] = None
    ):
        return sorted(
            [
                TorrentSearchResult(
                    provider=cls.provider_name(),
                    is_media=True,
                    imdb_id=result.get('imdb_id'),
                    type='movies',
                    title=result.get('title', '[No Title]')
                    + f' [movies][{lang}][{quality}]',
                    file=item.get('file'),
                    duration=int(result.get('runtime') or 0) * 60,
                    year=int(result.get('year') or 0),
                    description=result.get('synopsis'),
                    trailer=result.get('trailer'),
                    genres=result.get('genres', []),
                    image=result.get('images', {}).get('poster'),
                    rating=result.get('rating', {}).get('percentage'),
                    votes=result.get('rating', {}).get('votes'),
                    language=lang,
                    quality=quality,
                    size=item.get('size'),
                    seeds=item.get('seed'),
                    peers=item.get('peer'),
                    url=item.get('url'),
                )
                for result in results
                for (lang, items) in ((result or {}).get('torrents', {}) or {}).items()
                if not language or language == lang
                for (quality, item) in items.items()
                if result
            ],
            key=lambda item: item.seeds,
            reverse=True,
        )

    @classmethod
    def _results_to_tv_response(cls, results: List[dict]):
        return sorted(
            [
                TorrentSearchResult(
                    provider=cls.provider_name(),
                    is_media=True,
                    imdb_id=result.get('imdb_id'),
                    tvdb_id=result.get('tvdb_id'),
                    type='tv',
                    file=item.get('file'),
                    series=result.get('title'),
                    title=(
                        result.get('title', '[No Title]')
                        + f'[S{episode.get("season", 0):02d}E{episode.get("episode", 0):02d}] '
                        + f'{episode.get("title", "[No Title]")} [tv][{quality}]'
                    ),
                    duration=int(result.get('runtime') or 0) * 60,
                    year=int(result.get('year') or 0),
                    description=result.get('synopsis'),
                    overview=episode.get('overview'),
                    season=episode.get('season'),
                    episode=episode.get('episode'),
                    num_seasons=result.get('num_seasons'),
                    country=result.get('country'),
                    network=result.get('network'),
                    series_status=result.get('status'),
                    genres=result.get('genres', []),
                    image=result.get('images', {}).get('fanart'),
                    rating=result.get('rating', {}).get('percentage'),
                    votes=result.get('rating', {}).get('votes'),
                    quality=quality,
                    seeds=item.get('seeds'),
                    peers=item.get('peers'),
                    url=item.get('url'),
                )
                for result in results
                for episode in (result or {}).get('episodes', [])
                for quality, item in (episode.get('torrents', {}) or {}).items()
                if quality != '0'
            ],
            key=lambda item: (  # type: ignore
                '.'.join(
                    [  # type: ignore
                        (item.series or ''),
                        (item.quality or ''),
                        str(item.season or 0).zfill(2),
                        str(item.episode or 0).zfill(2),
                    ]
                )
            ),
        )

    def search_movies(self, query, language=None, **_):
        return self._results_to_movies_response(
            self._search_torrents(query, 'movies'), language=language
        )

    def search_tv(self, query, **_):
        return self._results_to_tv_response(self._search_torrents(query, 'tv'))

    categories = {
        'movies': search_movies,
        'tv': search_tv,
    }


# vim:sw=4:ts=4:et:



from typing import List, Optional, Dict

from platypush.common.sensors import Numeric
from platypush.entities.devices import Device
from platypush.entities.humidity import HumiditySensor
from platypush.entities.temperature import TemperatureSensor
from platypush.plugins import action
from platypush.plugins.sensor import SensorPlugin


# pylint: disable=too-many-ancestors
class SensorDhtPlugin(SensorPlugin):
    """
    Plugin to interact with a DHT11/DHT22/AM2302 temperature/humidity sensor through GPIO.
    """

    def __init__(
        self,
        sensor_type: str,
        pin: int,
        retries: int = 5,
        retry_seconds: int = 2,
        poll_interval: float = 5.0,
        **kwargs,
    ):
        """
        :param sensor_type: Type of sensor to be used (supported types: DHT11, DHT22, AM2302).
        :param pin: GPIO PIN where the sensor is connected.
        :param retries: Number of retries in case of failed read (default: 5).
        :param retry_seconds: Number of seconds to wait between retries (default: 2).
        """
        super().__init__(poll_interval=poll_interval, **kwargs)
        self.sensor_type = self._get_sensor_type(sensor_type)
        self.pin = pin
        self.retries = retries
        self.retry_seconds = retry_seconds

    @staticmethod
    def _get_sensor_type(sensor_type: str) -> int:
        import Adafruit_DHT

        sensor_type = sensor_type.upper()
        assert hasattr(
            Adafruit_DHT, sensor_type
        ), f'Unknown sensor type: {sensor_type}. Supported types: DHT11, DHT22, AM2302'

        return getattr(Adafruit_DHT, sensor_type)

    @action
    def read(
        self,
        sensor_type: Optional[str] = None,
        pin: Optional[int] = None,
        retries: Optional[int] = None,
        retry_seconds: Optional[int] = None,
        **__,
    ) -> Dict[str, float]:
        """
        Read data from the sensor.

        :param sensor_type: Default ``sensor_type`` override.
        :param pin: Default ``pin`` override.
        :param retries: Default ``retries`` override.
        :param retry_seconds: Default ``retry_seconds`` override.
        :return: A mapping with the measured temperature and humidity. Example:

            .. code-block:: json

                {
                    "humidity": 30.0,
                    "temperature": 25.5
                }

        """
        import Adafruit_DHT

        sensor_type = (  # type: ignore
            self._get_sensor_type(sensor_type) if sensor_type else self.sensor_type
        )
        pin = pin or self.pin
        retries = retries or self.retries
        retry_seconds = retry_seconds or self.retry_seconds
        humidity, temperature = Adafruit_DHT.read_retry(
            sensor=sensor_type, pin=pin, retries=retries, delay_seconds=retry_seconds
        )

        return {
            'humidity': humidity,
            'temperature': temperature,
        }

    @action
    def get_measurement(self, *_, **__) -> Dict[str, float]:
        """
        Get data from the sensor.

        :return: A mapping with the measured temperature and humidity. Example:

            .. code-block:: json

                {
                    "humidity": 30.0,
                    "temperature": 25.5
                }

        """
        return self.read()  # type: ignore

    def transform_entities(self, entities: Dict[str, Numeric]) -> List[Device]:
        return [
            Device(
                id='dht',
                name='DHT Sensor',
                children=[
                    TemperatureSensor(
                        id='dht:temperature',
                        name='temperature',
                        value=entities['temperature'],
                        unit='°C',
                    ),
                    HumiditySensor(
                        id='dht:humidity',
                        name='humidity',
                        value=entities['humidity'],
                        unit='%',
                    ),
                ],
            )
        ]


# vim:sw=4:ts=4:et:



from typing import Iterable, Optional, Type

import requests
from marshmallow import Schema

from platypush.plugins import Plugin, action
from platypush.schemas.media.jellyfin import (
    JellyfinArtistSchema,
    JellyfinCollectionSchema,
    JellyfinMovieSchema,
    JellyfinVideoSchema,
    JellyfinEpisodeSchema,
)


class MediaJellyfinPlugin(Plugin):
    """
    Plugin to interact with a Jellyfin media server.

    Note: As of February 2022, this plugin also works with Emby
    media server instances. Future back-compatibility if the two
    APIs diverge, however, is not guaranteed.
    """

    # Maximum number of results returned per query action
    _default_limit = 100

    def __init__(
        self, server: str, api_key: str, username: Optional[str] = None, **kwargs
    ):
        """
        :param server: Jellyfin base server URL (including ``http://`` or ``https://``).
        :param api_key: Server API key. You can generate one from
            ``http(s)://your-server/web/index.html#!/apikeys.html``.
        :param username: Customize results for the specified user
            (default: user associated to the API token if it's a user token, or the first created
            user on the platform).
        """
        super().__init__(**kwargs)
        self.server = server.rstrip('/')
        self.username = username
        self._api_key = api_key
        self.__user_id = None

    def _execute(self, method: str, url: str, *args, **kwargs) -> dict:
        url = '/' + url.lstrip('/')
        url = self.server + url

        kwargs['headers'] = {
            **kwargs.get('headers', {}),
            'X-Emby-Authorization': 'MediaBrowser Client="Platypush", Device="Platypush", '
            f'Token="{self._api_key}"',
        }

        rs = getattr(requests, method.lower())(url, *args, **kwargs)
        rs.raise_for_status()

        return rs.json()

    @property
    def _user_id(self) -> str:
        if not self.__user_id:
            try:
                self.__user_id = self._execute('GET', '/Users/Me')['Id']
            except requests.exceptions.HTTPError as e:
                assert (
                    e.response.status_code == 400
                ), f'Could not get the current user: {e}'

                self.__user_id = self._execute('GET', '/Users')[0]['Id']

        return self.__user_id

    def _query(
        self,
        url: str,
        schema_class: Optional[Type[Schema]] = None,
        query: Optional[str] = None,
        limit: Optional[int] = _default_limit,
        offset: int = 0,
        parent_id: Optional[str] = None,
        is_played: Optional[bool] = None,
        is_favourite: Optional[bool] = None,
        is_liked: Optional[bool] = None,
        genres: Optional[Iterable[str]] = None,
        tags: Optional[Iterable[str]] = None,
        years: Optional[Iterable[int]] = None,
        **kwargs,
    ) -> Iterable[dict]:
        filters = []
        if is_played is not None:
            filters.append('IsPlayed' if is_played else 'IsUnplayed')
        if is_liked is not None:
            filters.append('Likes' if is_liked else 'Dislikes')

        kwargs['params'] = {
            **({'isFavorite': is_favourite} if is_favourite is not None else {}),
            **({'searchTerm': query} if query else {}),
            **({'limit': limit} if limit else {}),
            'startIndex': offset,
            'includeMedia': True,
            'includeOverview': True,
            'recursive': True,
            **({'parentId': parent_id} if parent_id else {}),
            **({'genres': '|'.join(genres)} if genres else {}),
            **({'tags': '|'.join(tags)} if tags else {}),
            **({'years': ','.join(map(str, years))} if years else {}),
            **kwargs.get('params', {}),
        }

        results = self._execute(method='get', url=url, **kwargs).get('Items', [])
        if schema_class:
            results = schema_class().dump(results, many=True)

        return results

    def _flatten_series_result(self, search_result: dict) -> Iterable[dict]:
        episodes = []
        show_id = search_result['Id']
        seasons = self._execute(
            'get',
            f'/Shows/{show_id}/Seasons',
            params={
                'userId': self._user_id,
            },
        ).get('Items', [])

        for i, season in enumerate(seasons):
            episodes.extend(
                JellyfinEpisodeSchema().dump(
                    [
                        {**episode, 'SeasonIndex': i + 1}
                        for episode in self._execute(
                            'get',
                            f'/Shows/{show_id}/Episodes',
                            params={
                                'userId': self._user_id,
                                'seasonId': season['Id'],
                            },
                        ).get('Items', [])
                    ],
                    many=True,
                )
            )

        return episodes

    def _serialize_search_results(
        self, search_results: Iterable[dict]
    ) -> Iterable[dict]:
        serialized_results = []
        for result in search_results:
            if result['Type'] == 'CollectionFolder':
                result = JellyfinCollectionSchema().dump(result)
                result['type'] = 'collection'  # type: ignore
            elif result['Type'] == 'Movie':
                result = JellyfinMovieSchema().dump(result)
                result['type'] = 'movie'  # type: ignore
            elif result['Type'] == 'Video':
                result = JellyfinVideoSchema().dump(result)
                result['type'] = 'video'  # type: ignore
            elif result['Type'] == 'Series':
                serialized_results += self._flatten_series_result(result)
                for r in serialized_results:
                    r['type'] = 'episode'

            if isinstance(result, dict) and result.get('type'):
                serialized_results.append(result)

        return serialized_results

    @action
    def get_artists(
        self,
        limit: Optional[int] = _default_limit,
        offset: int = 0,
        query: Optional[str] = None,
        is_played: Optional[bool] = None,
        is_favourite: Optional[bool] = None,
        is_liked: Optional[bool] = None,
        genres: Optional[Iterable[str]] = None,
        tags: Optional[Iterable[str]] = None,
        years: Optional[Iterable[int]] = None,
    ) -> Iterable[dict]:
        """
        Get a list of artists on the server.

        :param limit: Maximum number of items to return (default: 100).
        :param offset: Return results starting from this (0-based) index (default: 0).
        :param query: Filter items by this term.
        :param is_played: Return only played items (or unplayed if set to False).
        :param is_liked: Return only liked items (or not liked if set to False).
        :param is_favourite: Return only favourite items (or not favourite if set to False).
        :param genres: Filter results by (a list of) genres.
        :param tags: Filter results by (a list of) tags.
        :param years: Filter results by (a list of) years.
        :return: .. schema:: media.jellyfin.JellyfinArtistSchema(many=True)
        """
        return self._query(
            '/Artists',
            schema_class=JellyfinArtistSchema,
            limit=limit,
            offset=offset,
            is_favourite=is_favourite,
            is_played=is_played,
            is_liked=is_liked,
            genres=genres,
            query=query,
            tags=tags,
            years=years,
        )

    @action
    def get_collections(self) -> Iterable[dict]:
        """
        Get the list of collections associated to the user on the server (Movies, Series, Channels etc.)

        :return: .. schema:: media.jellyfin.JellyfinCollectionSchema(many=True)
        """
        return self._query(
            f'/Users/{self._user_id}/Items',
            parent_id=None,
            schema_class=JellyfinCollectionSchema,
            params={'recursive': False},
        )

    @action
    def search(
        self,
        limit: Optional[int] = _default_limit,
        offset: int = 0,
        sort_desc: Optional[bool] = None,
        query: Optional[str] = None,
        collection: Optional[str] = None,
        parent_id: Optional[str] = None,
        has_subtitles: Optional[bool] = None,
        minimum_community_rating: Optional[int] = None,
        minimum_critic_rating: Optional[int] = None,
        is_played: Optional[bool] = None,
        is_favourite: Optional[bool] = None,
        is_liked: Optional[bool] = None,
        genres: Optional[Iterable[str]] = None,
        tags: Optional[Iterable[str]] = None,
        years: Optional[Iterable[int]] = None,
    ) -> Iterable[dict]:
        """
        Perform a search on the server.

        :param limit: Maximum number of items to return (default: 100).
        :param offset: Return results starting from this (0-based) index (default: 0).
        :param sort_desc: Return results in descending order if true, ascending if false.
        :param query: Filter items by this term.
        :param collection: ID/name of the collection to search (Movies, TV, Channels etc.)
        :param parent_id: Filter items under the specified parent ID.
        :param has_subtitles: Filter items with/without subtitles.
        :param minimum_community_rating: Filter by minimum community rating.
        :param minimum_critic_rating: Filter by minimum critic rating.
        :param is_played: Return only played items (or unplayed if set to False).
        :param is_liked: Return only liked items (or not liked if set to False).
        :param is_favourite: Return only favourite items (or not favourite if set to False).
        :param genres: Filter results by (a list of) genres.
        :param tags: Filter results by (a list of) tags.
        :param years: Filter results by (a list of) years.
        :return: The list of matching results.

        Schema for artists:
            .. schema:: media.jellyfin.JellyfinArtistSchema

        Schema for collections:
            .. schema:: media.jellyfin.JellyfinCollectionSchema

        Schema for movies:
            .. schema:: media.jellyfin.JellyfinMovieSchema

        Schema for episodes:
            .. schema:: media.jellyfin.JellyfinEpisodeSchema

        """
        if collection:
            collections = self.get_collections().output  # type: ignore
            matching_collections = [
                c
                for c in collections
                if c['id'] == collection or c['name'].lower() == collection.lower()
            ]

            if not matching_collections:
                return []  # No matching collections

            if not parent_id:
                parent_id = matching_collections[0]['id']

        results = self._query(
            f'/Users/{self._user_id}/Items',
            limit=limit,
            offset=offset,
            is_favourite=is_favourite,
            is_played=is_played,
            is_liked=is_liked,
            genres=genres,
            query=query,
            tags=tags,
            years=years,
            parent_id=parent_id,
            params={
                **(
                    {'sortOrder': 'Descending' if sort_desc else 'Ascending'}
                    if sort_desc is not None
                    else {}
                ),
                **(
                    {'hasSubtitles': has_subtitles} if has_subtitles is not None else {}
                ),
                **(
                    {'minCriticRating': minimum_critic_rating}
                    if minimum_critic_rating is not None
                    else {}
                ),
                **(
                    {'minCommunityRating': minimum_community_rating}
                    if minimum_community_rating is not None
                    else {}
                ),
            },
        )

        return self._serialize_search_results(results)



from contextlib import contextmanager
import logging
import pathlib
import os
import signal
import subprocess
import sys
from textwrap import dedent
from typing import Optional, Sequence

from platypush.bus import Bus
from platypush.bus.redis import RedisBus
from platypush.cli import parse_cmdline
from platypush.commands import CommandStream
from platypush.config import Config
from platypush.context import register_backends, register_plugins
from platypush.cron.scheduler import CronScheduler
from platypush.entities import init_entities_engine, EntitiesEngine
from platypush.event.processor import EventProcessor
from platypush.logger import Logger
from platypush.message.event import Event
from platypush.message.event.application import ApplicationStartedEvent
from platypush.message.request import Request
from platypush.message.response import Response
from platypush.utils import get_enabled_plugins, get_redis_conf

log = logging.getLogger('platypush')


class Application:
    """Main class for the Platypush application."""

    # Default Redis port
    _default_redis_port = 6379

    # backend_name => backend_obj map
    backends = None

    # number of executions retries before a request fails
    n_tries = 2

    def __init__(
        self,
        config_file: Optional[str] = None,
        workdir: Optional[str] = None,
        db: Optional[str] = None,
        logsdir: Optional[str] = None,
        cachedir: Optional[str] = None,
        device_id: Optional[str] = None,
        pidfile: Optional[str] = None,
        requests_to_process: Optional[int] = None,
        no_capture_stdout: bool = False,
        no_capture_stderr: bool = False,
        redis_queue: Optional[str] = None,
        verbose: bool = False,
        start_redis: bool = False,
        redis_host: Optional[str] = None,
        redis_port: Optional[int] = None,
        ctrl_sock: Optional[str] = None,
    ):
        """
        :param config_file: Configuration file. The order of precedence is:

            - The ``config_file`` parameter (or the ``-c``/``--config`` command
              line argument).
            - The ``PLATYPUSH_CONFIG`` environment variable.
            - ``./config.yaml``
            - ``~/.config/platypush/config.yaml``
            - ``/etc/platypush/config.yaml``

        :param workdir: Working directory where the application will store its
            data and integration plugins will store their data. The order of
            precedence is:

                - The ``workdir`` parameter (or the ``-w``/``--workdir`` command
                  line argument).
                - The ``PLATYPUSH_WORKDIR`` environment variable.
                - The ``workdir`` field in the configuration file.
                - ``~/.local/share/platypush``

        :param db: Main database engine for the application. Supports SQLAlchemy
            engine strings. The order of precedence is:

                - The ``db`` parameter (or the ``--main-db``/``--db`` command
                  line argument).
                - The ``PLATYPUSH_DB`` environment variable.
                - The ``db`` field in the configuration file.
                - ``sqlite:///<workdir>/main.db``

        :param logsdir: Logs directory where the application will store its logs.
            The order of precedence is:

                - The ``logsdir`` parameter (or the ``-l``/``--logsdir`` command
                  line argument).
                - The ``PLATYPUSH_LOGSDIR`` environment variable.
                - The ``logging`` -> ``filename`` field in the configuration
                  file (the ``format`` and ``level`` fields can be set as well
                  using Python logging configuration).
                - stdout and stderr

        :param cachedir: Directory where the application and the plugins will store
            their cache data. The order of precedence is:

                - The ``cachedir`` parameter (or the ``--cachedir`` command line
                  argument).
                - The ``PLATYPUSH_CACHEDIR`` environment variable.
                - The ``cachedir`` field in the configuration file.
                - ``~/.cache/platypush``

        :param device_id: Device ID used to identify this instance. The order
            of precedence is:

              - The ``device_id`` parameter (or the ``--device-id`` command
                line argument).
              - The ``PLATYPUSH_DEVICE_ID`` environment variable.
              - The ``device_id`` field in the configuration file.
              - The hostname of the machine.

        :param pidfile: File where platypush will store its PID upon launch,
           useful if you're planning to integrate the application within a
           service or a launcher script. Order of precedence:

               - The ``pidfile`` parameter (or the ``-P``/``--pidfile`` command
                 line argument).
               - The ``PLATYPUSH_PIDFILE`` environment variable.
               - No PID file.

        :param requests_to_process: Exit after processing the specified
            number of requests (default: None, loop forever). This is usually
            useful for testing purposes.
        :param no_capture_stdout: Set to true if you want to disable the
            stdout capture by the logging system (default: False).
        :param no_capture_stderr: Set to true if you want to disable the
            stderr capture by the logging system (default: False).
        :param redis_queue: Name of the (Redis) queue used for dispatching
            messages. Order of precedence:

                - The ``redis_queue`` parameter (or the ``--redis-queue`` command
                  line argument).
                - The ``PLATYPUSH_REDIS_QUEUE`` environment variable.
                - ``platypush/bus``

        :param verbose: Enable debug/verbose logging, overriding the stored
            configuration (default: False).
        :param start_redis: If set, it starts a managed Redis instance upon
            boot (it requires the ``redis-server`` executable installed on the
            server). This is particularly useful when running the application
            inside of Docker containers, without relying on ``docker-compose``
            to start multiple containers, and in tests (default: False).
        :param redis_host: Host of the Redis server to be used. The order of
            precedence is:

                - The ``redis_host`` parameter (or the ``--redis-host`` command
                  line argument).
                - The ``PLATYPUSH_REDIS_HOST`` environment variable.
                - The ``redis`` -> ``host`` field in the configuration file.
                - The ``backend.redis`` -> ``redis_args`` -> ``host`` field in
                  the configuration file.
                - ``localhost``

        :param redis_port: Port of the Redis server to be used. The order of
            precedence is:

                - The ``redis_port`` parameter (or the ``--redis-port`` command
                  line argument).
                - The ``PLATYPUSH_REDIS_PORT`` environment variable.
                - The ``redis`` -> ``port`` field in the configuration file.
                - The ``backend.redis`` -> ``redis_args`` -> ``port`` field in
                  the configuration file.
                - ``6379``

        :param ctrl_sock: If set, it identifies a path to a UNIX domain socket
            that the application can use to send control messages (e.g. STOP
            and RESTART) to its parent.
        """

        self.pidfile = pidfile or os.environ.get('PLATYPUSH_PIDFILE')
        self.bus: Optional[Bus] = None
        self.redis_queue = (
            redis_queue
            or os.environ.get('PLATYPUSH_REDIS_QUEUE')
            or RedisBus.DEFAULT_REDIS_QUEUE
        )

        os.environ['PLATYPUSH_REDIS_QUEUE'] = self.redis_queue
        self.config_file = config_file or os.environ.get('PLATYPUSH_CONFIG')
        self.verbose = verbose
        self.db_engine = db or os.environ.get('PLATYPUSH_DB')
        self.device_id = device_id or os.environ.get('PLATYPUSH_DEVICE_ID')

        self.logsdir = self.expand_path(logsdir or os.environ.get('PLATYPUSH_LOGSDIR'))
        self.workdir = self.expand_path(workdir or os.environ.get('PLATYPUSH_WORKDIR'))
        self.cachedir = self.expand_path(
            cachedir or os.environ.get('PLATYPUSH_CACHEDIR')
        )
        Config.init(
            self.config_file,
            device_id=self.device_id,
            workdir=self.workdir,
            db=self.db_engine,
            cachedir=self.cachedir,
            ctrl_sock=self.expand_path(ctrl_sock),
        )

        self.no_capture_stdout = no_capture_stdout
        self.no_capture_stderr = no_capture_stderr
        self.event_processor = EventProcessor()
        self.entities_engine: Optional[EntitiesEngine] = None
        self.requests_to_process = requests_to_process
        self.processed_requests = 0
        self.cron_scheduler = None
        self.start_redis = start_redis
        self.redis_host = redis_host or os.environ.get('PLATYPUSH_REDIS_HOST')
        self.redis_port = redis_port or os.environ.get('PLATYPUSH_REDIS_PORT')
        self._redis_conf = {
            'host': self.redis_host or 'localhost',
            'port': self.redis_port or self._default_redis_port,
        }

        self._redis_proc: Optional[subprocess.Popen] = None
        self.cmd_stream = CommandStream(ctrl_sock)

        self._init_bus()
        self._init_logging()

    @staticmethod
    def expand_path(path: Optional[str]) -> Optional[str]:
        return os.path.abspath(os.path.expanduser(path)) if path else None

    def _init_bus(self):
        self._redis_conf = {**self._redis_conf, **get_redis_conf()}
        Config.set('redis', self._redis_conf)
        self.bus = RedisBus(
            redis_queue=self.redis_queue,
            on_message=self.on_message(),
            **self._redis_conf,
        )

    def _init_logging(self):
        logging_conf = Config.get('logging') or {}
        if self.verbose:
            logging_conf['level'] = logging.DEBUG

        if self.logsdir:
            logging_conf['filename'] = os.path.join(self.logsdir, 'platypush.log')
            logging_conf.pop('stream', None)

        if logging_conf.get('filename'):
            pathlib.Path(os.path.dirname(logging_conf['filename'])).mkdir(
                parents=True, exist_ok=True
            )

        Config.set('logging', logging_conf)
        logging.basicConfig(**logging_conf)

    def _start_redis(self):
        if self._redis_proc and self._redis_proc.poll() is None:
            log.warning(
                'A local Redis instance is already running, refusing to start it again'
            )
            return

        port = self._redis_conf['port']
        log.info('Starting local Redis instance on %s', port)
        redis_cmd_args = [
            'redis-server',
            '--bind',
            'localhost',
            '--port',
            str(port),
        ]

        try:
            self._redis_proc = subprocess.Popen(  # pylint: disable=consider-using-with
                redis_cmd_args,
                stdout=subprocess.PIPE,
            )
        except Exception as e:
            log.error(
                'Failed to start local Redis instance: "%s": %s',
                ' '.join(redis_cmd_args),
                e,
            )

            sys.exit(1)

        log.info('Waiting for Redis to start')
        for line in self._redis_proc.stdout:  # type: ignore
            if b'Ready to accept connections' in line:
                break

    def _stop_redis(self):
        if self._redis_proc and self._redis_proc.poll() is None:
            log.info('Stopping local Redis instance')
            self._redis_proc.kill()
            self._redis_proc = None

    @classmethod
    def from_cmdline(cls, args: Sequence[str]) -> "Application":
        """
        Build the app from command line arguments.
        """
        opts = parse_cmdline(args)
        return cls(
            config_file=opts.config,
            workdir=opts.workdir,
            cachedir=opts.cachedir,
            logsdir=opts.logsdir,
            db=opts.db_engine,
            device_id=opts.device_id,
            pidfile=opts.pidfile,
            no_capture_stdout=opts.no_capture_stdout,
            no_capture_stderr=opts.no_capture_stderr,
            redis_queue=opts.redis_queue,
            verbose=opts.verbose,
            start_redis=opts.start_redis,
            redis_host=opts.redis_host,
            redis_port=opts.redis_port,
            ctrl_sock=opts.ctrl_sock,
        )

    def on_message(self):
        """
        Default message handler.
        """

        def _f(msg):
            """
            on_message closure
            Params:
                msg -- platypush.message.Message instance
            """

            if isinstance(msg, Request):
                try:
                    msg.execute(n_tries=self.n_tries)
                except PermissionError:
                    log.info('Dropped unauthorized request: %s', msg)

                self.processed_requests += 1
                if (
                    self.requests_to_process
                    and self.processed_requests >= self.requests_to_process
                ):
                    self.stop()
            elif isinstance(msg, Response):
                msg.log()
            elif isinstance(msg, Event):
                msg.log()
                self.event_processor.process_event(msg)

        return _f

    def stop(self):
        """Stops the backends and the bus."""
        from platypush.plugins import RunnablePlugin

        log.info('Stopping the application')
        backends = (self.backends or {}).copy().values()
        runnable_plugins = [
            plugin
            for plugin in get_enabled_plugins().values()
            if isinstance(plugin, RunnablePlugin)
        ]

        for backend in backends:
            backend.stop()

        for plugin in runnable_plugins:
            # This is required because some plugins may redefine the `stop` method.
            # In that case, at the very least the _should_stop event should be
            # set to notify the plugin to stop.
            plugin._should_stop.set()  # pylint: disable=protected-access
            plugin.stop()

        for backend in backends:
            backend.wait_stop()

        for plugin in runnable_plugins:
            plugin.wait_stop()

        if self.entities_engine:
            self.entities_engine.stop()
            self.entities_engine.wait_stop()
            self.entities_engine = None

        if self.cron_scheduler:
            self.cron_scheduler.stop()
            self.cron_scheduler.wait_stop()
            self.cron_scheduler = None

        if self.bus:
            self.bus.stop()
            self.bus = None

        if self.start_redis:
            self._stop_redis()

        log.info('Exiting application')

    @contextmanager
    def _open_pidfile(self):
        if self.pidfile:
            try:
                with open(self.pidfile, 'w') as f:
                    f.write(str(os.getpid()))
            except OSError as e:
                log.warning('Failed to create PID file %s: %s', self.pidfile, e)

        yield

        if self.pidfile:
            try:
                os.remove(self.pidfile)
            except OSError as e:
                log.warning('Failed to remove PID file %s: %s', self.pidfile, e)

    def _run(self):
        from platypush import __version__

        if not self.no_capture_stdout:
            sys.stdout = Logger(log.info)
        if not self.no_capture_stderr:
            sys.stderr = Logger(log.warning)

        log.info(
            dedent(
                r'''
                  _____  _       _                         _
                 |  __ \| |     | |                       | |
                 | |__) | | __ _| |_ _   _ _ __  _   _ ___| |__
                 |  ___/| |/ _` | __| | | | '_ \| | | / __| '_ \
                 | |    | | (_| | |_| |_| | |_) | |_| \__ \ | | |
                 |_|    |_|\__,_|\__|\__, | .__/ \__,_|___/_| |_|
                                      __/ | |
                                     |___/|_|
                        '''
            )
        )
        log.info('---- Starting Platypush v.%s', __version__)

        # Start the local Redis service if required
        if self.start_redis:
            self._start_redis()

        # Initialize the backends and link them to the bus
        self.backends = register_backends(bus=self.bus, global_scope=True)

        # Start the backend threads
        for backend in self.backends.values():
            backend.start()

        # Initialize the plugins
        register_plugins(bus=self.bus)

        # Initialize the entities engine
        self.entities_engine = init_entities_engine()

        # Start the cron scheduler
        if Config.get_cronjobs():
            self.cron_scheduler = CronScheduler(jobs=Config.get_cronjobs())
            self.cron_scheduler.start()

        assert self.bus, 'The bus is not running'
        self.bus.post(ApplicationStartedEvent())

        # Poll for messages on the bus
        try:
            self.bus.poll()
        except KeyboardInterrupt:
            log.info('SIGINT received, terminating application')
            # Ignore other SIGINT signals
            signal.signal(signal.SIGINT, signal.SIG_IGN)
        finally:
            self.stop()

    def run(self):
        """Run the application."""

        with self._open_pidfile():
            self._run()


app: Optional[Application] = None


def main(*args: str):
    """
    Application entry point.
    """
    global app

    app = Application.from_cmdline(args)

    try:
        app.run()
    except KeyboardInterrupt:
        pass

    return 0


# vim:sw=4:ts=4:et:



from collections import defaultdict
from typing import Type

from .._base import XmppBasePlugin
from ._base import XmppBaseHandler


class XmppHandlersRegistry(defaultdict):
    """
    A registry of the initialized XMPP handlers.
    """

    def __init__(self, plugin: XmppBasePlugin):
        super().__init__()
        self._plugin = plugin

    def __missing__(self, hndl_type: Type[XmppBaseHandler]) -> XmppBaseHandler:
        return self._plugin.register_handler(hndl_type)



from dataclasses import dataclass, field

from platypush.schemas.dataclasses import percent_field


@dataclass
class MemoryStats:
    """
    Memory stats data class.
    """

    total: int = field(
        metadata={
            'metadata': {
                'description': 'Total available memory, in bytes',
            }
        }
    )

    available: int = field(
        metadata={
            'metadata': {
                'description': 'Available memory, in bytes',
            }
        }
    )

    used: int = field(
        metadata={
            'metadata': {
                'description': 'Used memory, in bytes',
            }
        }
    )

    free: int = field(
        metadata={
            'metadata': {
                'description': 'Free memory, in bytes',
            }
        }
    )

    active: int = field(
        metadata={
            'metadata': {
                'description': 'Size of the active memory, in bytes',
            }
        }
    )

    inactive: int = field(
        metadata={
            'metadata': {
                'description': 'Size of the inactive memory, in bytes',
            }
        }
    )

    buffers: int = field(
        metadata={
            'metadata': {
                'description': 'Size of the buffered memory, in bytes',
            }
        }
    )

    cached: int = field(
        metadata={
            'metadata': {
                'description': 'Size of the cached memory, in bytes',
            }
        }
    )

    shared: int = field(
        metadata={
            'metadata': {
                'description': 'Size of the shared memory, in bytes',
            }
        }
    )

    percent: float = percent_field()


@dataclass
class SwapStats:
    """
    Swap memory stats data class.
    """

    total: int = field(
        metadata={
            'metadata': {
                'description': 'Total available memory, in bytes',
            }
        }
    )

    used: int = field(
        metadata={
            'metadata': {
                'description': 'Used memory, in bytes',
            }
        }
    )

    free: int = field(
        metadata={
            'metadata': {
                'description': 'Free memory, in bytes',
            }
        }
    )

    percent: float = percent_field()



import logging

from sqlalchemy import Column, ForeignKey, Integer, String

from platypush.common.db import is_defined

from . import Entity

logger = logging.getLogger(__name__)


if not is_defined('variable'):

    class Variable(Entity):
        """
        Models a variable entity.
        """

        __tablename__ = 'variable'

        id = Column(
            Integer, ForeignKey('entity.id', ondelete='CASCADE'), primary_key=True
        )
        value = Column(String)

        __table_args__ = {'keep_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from typing import Optional, Union, Sequence

import requests
from requests.auth import HTTPBasicAuth

from platypush.plugins import Plugin, action


class MailgunPlugin(Plugin):
    """
    Mailgun integration.
    """

    def __init__(
        self,
        api_key: str,
        api_base_url: str = 'https://api.mailgun.net/v3',
        domain: Optional[str] = None,
        timeout: float = 20.0,
        **kwargs,
    ):
        """
        :param api_key: Mailgun API secret key.
        :param api_base_url: Use ``https://api.eu.mailgun.net/v3`` if you are using an EU account.
        :param domain: Default registered domain that should be used for sending
            emails if not specified in the :meth:`.send` action.
        :param timeout: Default timeout for the requests (default: 20 seconds).
        """
        super().__init__(**kwargs)
        self._api_key = api_key
        self._api_base_url = api_base_url
        self._domain = domain
        self._timeout = timeout

    @action
    def send(
        self,
        to: Union[str, Sequence[str]],
        from_: Optional[str] = None,
        cc: Optional[Union[str, Sequence[str]]] = None,
        bcc: Optional[Union[str, Sequence[str]]] = None,
        subject: str = '',
        body: str = '',
        body_type: str = 'plain',
        domain: Optional[str] = None,
        **kwargs,
    ):
        """
        Send an email through Mailgun.

        :param domain: From which registered domain the email(s) should be sent
            (default: the domain specified in the plugin configuration).
        :param to: Receiver(s), as comma-separated strings or list.
        :param from_: Sender email address (``from`` is also supported outside of Python contexts).
        :param cc: Carbon-copy addresses, as comma-separated strings or list
        :param bcc: Blind carbon-copy addresses, as comma-separated strings or list
        :param subject: Mail subject.
        :param body: Mail body.
        :param body_type: Mail body type - ``text`` or ``html``.
        """
        domain = domain or self._domain
        assert domain, 'No domain specified'
        from_ = from_ or kwargs.pop('from', None)
        rs = requests.post(
            f'{self._api_base_url}/{domain}/messages',
            timeout=self._timeout,
            auth=HTTPBasicAuth('api', self._api_key),
            data={
                'to': ', '.join([to] if isinstance(to, str) else to),
                'subject': subject,
                **{'html' if body_type == 'html' else 'text': body},
                **({'from': from_} if from_ else {}),
                **(
                    {'cc': ', '.join([cc] if isinstance(cc, str) else cc)} if cc else {}
                ),
                **(
                    {'bcc': ', '.join([bcc] if isinstance(bcc, str) else bcc)}
                    if bcc
                    else {}
                ),
            },
        )

        rs.raise_for_status()


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from dataclasses import asdict, is_dataclass
import decimal
import datetime
from enum import Enum
import io
import logging
import inspect
import json
import time
from typing import Union
from uuid import UUID

_logger = logging.getLogger('platypush')


class JSONAble(ABC):
    """
    Generic interface for JSON-able objects.
    """

    @abstractmethod
    def to_json(self) -> Union[str, list, dict]:
        raise NotImplementedError()


class Message:
    """
    Message generic class
    """

    class Encoder(json.JSONEncoder):
        """
        JSON encoder that can serialize custom types commonly handled in
        Platypush messages.
        """

        @staticmethod
        def parse_numpy(obj):
            try:
                import numpy as np

                if isinstance(obj, np.floating):
                    return float(obj)
                if isinstance(obj, np.integer):
                    return int(obj)
                if isinstance(obj, np.ndarray):
                    return obj.tolist()
                if isinstance(obj, decimal.Decimal):
                    return float(obj)
                if isinstance(obj, (bytes, bytearray)):
                    return '0x' + ''.join([f'{x:02x}' for x in obj])
                if callable(obj):
                    return f'<function at {obj.__module__}.{obj.__name__}>'
            except (AttributeError, ImportError, TypeError):
                pass

            return

        @staticmethod
        def parse_datetime(obj):
            if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
                return obj.isoformat()

        def default(self, o):
            from platypush.procedure import Procedure

            value = self.parse_datetime(o)
            if value is not None:
                return value

            if isinstance(o, set):
                return list(o)

            if isinstance(o, UUID):
                return str(o)

            value = self.parse_numpy(o)
            if value is not None:
                return value

            if isinstance(o, JSONAble):
                return o.to_json()

            if isinstance(o, Procedure):
                return o.to_dict()

            if isinstance(o, Enum):
                return o.value

            if isinstance(o, Exception):
                return f'<{o.__class__.__name__}>' + (f' {o}' if o else '')

            if is_dataclass(o):
                return asdict(o)

            if isinstance(o, Message):
                return o.to_dict(o)

            # Don't serialize I/O wrappers/objects
            if isinstance(o, io.IOBase):
                return None

            try:
                return super().default(o)
            except Exception as e:
                _logger.warning(
                    'Could not serialize object type %s: %s: %s', type(o), e, o
                )

    def __init__(self, *_, timestamp=None, logging_level=logging.INFO, **__):
        self.timestamp = timestamp or time.time()
        self.logging_level = logging_level
        self._logger = _logger
        self._default_log_prefix = ''

    def log(self, prefix=''):
        if self.logging_level is None:
            return  # Skip logging

        log_func = self._logger.info
        if self.logging_level == logging.DEBUG:
            log_func = self._logger.debug
        elif self.logging_level == logging.WARNING:
            log_func = self._logger.warning
        elif self.logging_level == logging.ERROR:
            log_func = self._logger.error
        elif self.logging_level == logging.FATAL:
            log_func = self._logger.fatal

        if not prefix:
            prefix = self._default_log_prefix
        log_func('%s%s', prefix, self)

    def __str__(self):
        """
        Overrides the str() operator and converts
        the message into a UTF-8 JSON string
        """

        return json.dumps(
            {
                attr: getattr(self, attr)
                for attr in self.__dir__()
                if (attr != '_timestamp' or not attr.startswith('_'))
                and not inspect.ismethod(getattr(self, attr))
            },
            cls=self.Encoder,
        ).replace('\n', ' ')

    def __bytes__(self):
        """
        Overrides the bytes() operator, converts the message into
        its JSON-serialized UTF-8-encoded representation
        """
        return str(self).encode('utf-8')

    @classmethod
    def parse(cls, msg):
        """
        Parse a generic message into a key-value dictionary

        :param msg: Original message. It can be a dictionary, a Message,
           or a string/bytearray, as long as it's valid UTF-8 JSON
        """

        if isinstance(msg, cls):
            msg = str(msg)
        if isinstance(msg, (bytes, bytearray)):
            msg = msg.decode('utf-8')
        if isinstance(msg, str):
            try:
                msg = json.loads(msg.strip())
            except (ValueError, TypeError):
                _logger.warning('Invalid JSON message: %s', msg)

        assert isinstance(msg, dict)

        if '_timestamp' not in msg:
            msg['_timestamp'] = time.time()

        return msg

    @classmethod
    def to_dict(cls, msg):
        """
        Converts a Message object into a dictionary

        :param msg: Message object
        """

        return {
            k: v
            for k, v in cls.parse(msg).items()
            if k not in ('id', 'token', 'target', 'origin', '_timestamp')
        }

    @classmethod
    def build(cls, msg):
        """
        Builds a Message object from a dictionary.

        :param msg: The message as a key-value dictionary, Message object or JSON string
        """
        from platypush.utils import get_message_class_by_type

        msg = cls.parse(msg)
        msgtype = get_message_class_by_type(msg['type'])
        if msgtype != cls:
            return msgtype.build(msg)


# vim:sw=4:ts=4:et:



from typing import List, Optional

import requests

from platypush.plugins.weather import WeatherPlugin
from platypush.schemas.weather.openweathermap import WeatherSchema


class WeatherOpenweathermapPlugin(WeatherPlugin):  # pylint: disable=too-many-ancestors
    """
    OpenWeatherMap plugin.

    You'll need an API token from `OpenWeatherMap
    <https://openweathermap.org/api>`_ in order to use this API.
    """

    base_url = 'https://api.openweathermap.org/data/2.5'

    def __init__(
        self,
        token: str,
        location: Optional[str] = None,
        city_id: Optional[int] = None,
        lat: Optional[float] = None,
        long: Optional[float] = None,
        zip_code: Optional[str] = None,
        units: str = 'metric',
        lang: Optional[str] = None,
        **kwargs,
    ):
        """
        :param token: OpenWeatherMap API token.
        :param location: If set, then this location will be used by default for
            weather lookup. If multiple locations share the same name you can
            disambiguate by specifying the country code as well - e.g.
            ``London,GB``.
        :param city_id: If set, then this city ID will be used by default for
            weather lookup. The full list of city IDs is available `here
            <https://bulk.openweathermap.org/sample/>`_.
        :param lat: If lat/long are set, then the weather by default will be
            retrieved for the specified geo location.
        :param long: If lat/long are set, then the weather by default will be
            retrieved for the specified geo location.
        :param zip_code: If set, then this ZIP code (should be in the form
            ``zip,country_code``) will be used by default for weather lookup.
        :param units: Supported: ``metric`` (default), ``standard`` and
            ``imperial``.
        :param lang: Language code for the weather description (default: en).
        """
        super().__init__(**kwargs)
        self._token = token
        self._location_query = None
        self._location_query = self._get_location_query(
            location=location, city_id=city_id, lat=lat, long=long, zip_code=zip_code
        )
        self.units = units
        self.lang = lang

    def _get_location_query(
        self,
        location: Optional[str] = None,
        city_id: Optional[int] = None,
        lat: Optional[float] = None,
        long: Optional[float] = None,
        zip_code: Optional[str] = None,
    ) -> dict:
        if city_id:
            return {'id': city_id}
        if lat and long:
            return {'lat': lat, 'lon': long}
        if zip_code:
            return {'zip': zip_code}
        if location:
            return {'q': location}

        assert self._location_query, 'Specify either location, city_id or lat/long'
        return self._location_query

    def _weather_request(
        self,
        path: str,
        *_,
        location: Optional[str] = None,
        city_id: Optional[int] = None,
        lat: Optional[float] = None,
        long: Optional[float] = None,
        zip_code: Optional[str] = None,
        units: Optional[str] = None,
        **__,
    ) -> dict:
        units = units or self.units
        params = {
            'appid': self._token,
            'units': units,
            'lang': self.lang,
            **self._get_location_query(
                location=location,
                city_id=city_id,
                lat=lat,
                long=long,
                zip_code=zip_code,
            ),
        }

        rs = requests.get(f'{self.base_url}/{path}', params=params, timeout=10)
        rs.raise_for_status()
        return rs.json()

    def _get_current_weather(
        self,
        *_,
        location: Optional[str] = None,
        city_id: Optional[int] = None,
        lat: Optional[float] = None,
        long: Optional[float] = None,
        zip_code: Optional[str] = None,
        units: Optional[str] = None,
        **__,
    ) -> dict:
        units = units or self.units
        return dict(
            WeatherSchema().dump(
                {
                    'units': units,
                    **self._weather_request(
                        'weather',
                        location=location,
                        city_id=city_id,
                        lat=lat,
                        long=long,
                        zip_code=zip_code,
                        units=units,
                    ),
                }
            )
        )

    def _get_forecast(
        self,
        *_,
        location: Optional[str] = None,
        city_id: Optional[int] = None,
        lat: Optional[float] = None,
        long: Optional[float] = None,
        zip_code: Optional[str] = None,
        units: Optional[str] = None,
        **__,
    ) -> List[dict]:
        units = units or self.units
        return list(
            WeatherSchema().dump(
                [
                    {
                        'units': units,
                        **data,
                    }
                    for data in self._weather_request(
                        'forecast',
                        location=location,
                        city_id=city_id,
                        lat=lat,
                        long=long,
                        zip_code=zip_code,
                        units=units,
                    ).get('list', [])
                ],
                many=True,
            )
        )



from dataclasses import dataclass
from typing import Optional, List, Union


@dataclass
class Pin:
    """
    This class models the configuration for the PIN of a device.
    """

    number: int
    name: Optional[str] = None
    pwm: bool = False
    pull_up: bool = False


@dataclass
class Device:
    """
    This class models the properties of a configured ESP device.
    """

    host: str
    port: int = 8266
    password: Optional[str] = None
    name: Optional[str] = None
    pins: Optional[List[Union[Pin, dict]]] = None

    @property
    def _pins(self):
        return [
            pin if isinstance(pin, Pin) else Pin(**pin) for pin in (self.pins or [])
        ]

    @property
    def _pins_by_name(self):
        return {pin.name: pin for pin in self._pins if pin.name}

    @property
    def _pins_by_number(self):
        return {pin.number: pin for pin in self._pins}

    def get_pin(self, pin) -> int:
        try:
            return int(pin)
        except ValueError:
            pass

        pin_obj = self._pins_by_name.get(pin)
        assert pin_obj, f'No such PIN configured: {pin}'
        return pin_obj.number


# vim:sw=4:ts=4:et:



from marshmallow import pre_load

from .._base import SystemBaseSchema


class MemoryStatsBaseSchema(SystemBaseSchema):
    """
    Base schema for memory stats.
    """

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)

        # Normalize the percentage between 0 and 1
        if data.get('percent') is not None:
            data['percent'] /= 100
        return data



import aioxmpp

from ._base import XmppBaseHandler


# pylint: disable=too-many-ancestors,too-few-public-methods
class XmppPingHandler(XmppBaseHandler):
    """
    Handler for the XMPP ping logic.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ping: aioxmpp.PingService = self._client.summon(aioxmpp.PingService)



from dataclasses import asdict
import multiprocessing
from threading import Timer
from typing import Optional, Type
from time import time

from platypush.context import get_bus
from platypush.message.event.joystick import (
    JoystickConnectedEvent,
    JoystickDisconnectedEvent,
    JoystickEvent,
    JoystickStateEvent,
)
from platypush.schemas.joystick import JoystickDeviceSchema

from ._inputs import GamePad, InputEvent, UnpluggedError
from ._state import ConnectedState, JoystickDeviceState, JoystickState


class JoystickManager(multiprocessing.Process):
    """
    A process that monitors and publishes joystick events.
    """

    MAX_TRIG_VAL = 1 << 8
    MAX_JOY_VAL = 1 << 15
    THROTTLE_INTERVAL = 0.2

    def __init__(
        self,
        device: GamePad,
        poll_interval: Optional[float],
        state_queue: multiprocessing.Queue,
    ):
        super().__init__()
        self.device = device
        self.poll_interval = poll_interval
        self.state = JoystickState()
        self._state_queue = state_queue
        self._connected_state = ConnectedState.UNKNOWN
        self._stop_evt = multiprocessing.Event()
        self._state_throttler: Optional[Timer] = None
        self._state_timestamp: float = 0

    @property
    def should_stop(self):
        return self._stop_evt.is_set()

    def wait_stop(self, timeout: Optional[float] = None):
        self._stop_evt.wait(timeout=timeout or self.poll_interval)

    def _enqueue_state(self):
        now = time()
        self._state_queue.put_nowait(
            JoystickDeviceState(
                device=self.device.get_char_device_path(), state=self.state
            )
        )

        if now - self._state_timestamp >= self.THROTTLE_INTERVAL:
            self._post_state()
            self._state_timestamp = now
            return

        self._state_timestamp = now
        self._state_throttler = Timer(self.THROTTLE_INTERVAL, self._post_state)
        self._state_throttler.start()

    def _stop_state_throrttler(self):
        if self._state_throttler:
            self._state_throttler.cancel()
            self._state_throttler = None

    def _post_state(self):
        self._post_event(JoystickStateEvent, state=asdict(self.state))
        self._stop_state_throrttler()

    def _parse_event(self, event: InputEvent):  # pylint: disable=too-many-branches
        """
        Solution adapted from https://stackoverflow.com/questions/46506850.
        """
        if event.code == "ABS_Y":
            # normalize between -1 and 1
            self.state.left_joystick_y = event.state / self.MAX_JOY_VAL
        elif event.code == "ABS_X":
            # normalize between -1 and 1
            self.state.left_joystick_x = event.state / self.MAX_JOY_VAL
        elif event.code == "ABS_RY":
            # normalize between -1 and 1
            self.state.right_joystick_y = event.state / self.MAX_JOY_VAL
        elif event.code == "ABS_RX":
            # normalize between -1 and 1
            self.state.right_joystick_x = event.state / self.MAX_JOY_VAL
        elif event.code == "ABS_Z":
            # normalize between 0 and 1
            self.state.left_trigger = event.state / self.MAX_TRIG_VAL
        elif event.code == "ABS_RZ":
            # normalize between 0 and 1
            self.state.right_trigger = event.state / self.MAX_TRIG_VAL
        elif event.code == "BTN_TL":
            self.state.left_bumper = event.state
        elif event.code == "BTN_TR":
            self.state.right_bumper = event.state
        elif event.code == "BTN_SOUTH":
            self.state.a = event.state
        elif event.code == "BTN_NORTH":
            # previously switched with X
            self.state.y = event.state
        elif event.code == "BTN_WEST":
            # previously switched with Y
            self.state.x = event.state
        elif event.code == "BTN_EAST":
            self.state.b = event.state
        elif event.code == "BTN_THUMBL":
            self.state.left_thumb = event.state
        elif event.code == "BTN_THUMBR":
            self.state.right_thumb = event.state
        elif event.code == "BTN_SELECT":
            self.state.back = event.state
        elif event.code == "BTN_START":
            self.state.start = event.state
        elif event.code == "BTN_TRIGGER_HAPPY1":
            self.state.left_dir_pad = event.state
        elif event.code == "BTN_TRIGGER_HAPPY2":
            self.state.right_dir_pad = event.state
        elif event.code == "BTN_TRIGGER_HAPPY3":
            self.state.up_dir_pad = event.state
        elif event.code == "BTN_TRIGGER_HAPPY4":
            self.state.down_dir_pad = event.state

    def _post_event(
        self, type: Type[JoystickEvent], **kwargs  # pylint: disable=redefined-builtin
    ):
        get_bus().post(
            type(device=dict(JoystickDeviceSchema().dump(self.device)), **kwargs)
        )

    def _on_connect(self):
        if self._connected_state != ConnectedState.CONNECTED:
            self._connected_state = ConnectedState.CONNECTED
            self._post_event(JoystickConnectedEvent)

    def _on_disconnect(self):
        if self._connected_state != ConnectedState.DISCONNECTED:
            self._connected_state = ConnectedState.DISCONNECTED
            self._post_event(JoystickDisconnectedEvent)

    def _loop(self):
        try:
            for event in self.device.read():
                self._on_connect()
                prev_state = asdict(self.state)
                self._parse_event(event)
                new_state = asdict(self.state)

                if prev_state != new_state:
                    self._enqueue_state()
        except (UnpluggedError, OSError):
            self._on_disconnect()
        finally:
            self.wait_stop(self.poll_interval)

    def run(self):
        try:
            while not self.should_stop:
                try:
                    self._loop()
                except KeyboardInterrupt:
                    break
        finally:
            self._on_disconnect()

    def stop(self):
        self._stop_evt.set()
        self._stop_state_throrttler()



from typing import Optional, Union

from platypush.message.event.xmpp import XmppDisconnectedEvent

from ._base import XmppBaseHandler


# pylint: disable=too-many-ancestors
class XmppConnectionHandler(XmppBaseHandler):
    """
    Handler for XMPP connection/disconnection events.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._client.on_failure.connect(self._on_disconnect())  # type: ignore
        self._client.on_stopped.connect(self._on_disconnect())  # type: ignore

    def _on_disconnect(self, reason: Optional[Union[str, Exception]] = None):
        def callback(*_, **__):
            if not self._state.disconnect_notified.is_set():
                self._post_event(XmppDisconnectedEvent, reason=reason)
            self._state.disconnect_notified.set()

        return callback

    def disconnect(self, reason: Optional[Union[str, Exception]] = None):
        self._on_disconnect(reason=reason)()



from contextlib import contextmanager
from dataclasses import dataclass
import json
import logging
from multiprocessing import RLock
from typing import Generator, Iterable, Optional, Set, Union

from redis import ConnectionError as RedisConnectionError
from redis.client import PubSub

from platypush.config import Config
from platypush.message import Message as AppMessage
from platypush.utils import get_redis

logger = logging.getLogger(__name__)

MessageType = Union[AppMessage, bytes, str, dict, list, set, tuple]
"""Types of supported messages on Redis/websocket channels."""


@dataclass
class Message:
    """
    A wrapper for a message received on a Redis subscription.
    """

    data: bytes
    """The data received in the message."""
    channel: str
    """The channel the message was received on."""


class PubSubMixin:
    """
    A mixin for Tornado route handlers that support pub/sub mechanisms.
    """

    def __init__(self, *_, subscriptions: Optional[Iterable[str]] = None, **__):
        self._pubsub: Optional[PubSub] = None
        """Pub/sub proxy."""
        self._subscriptions: Set[str] = set(subscriptions or [])
        """Set of current channel subscriptions."""
        self._pubsub_lock = RLock()
        """
        Subscriptions lock. It ensures that the list of subscriptions is
        manipulated by one thread or process at the time.
        """

        self.subscribe(*self._subscriptions)

    @property
    @contextmanager
    def pubsub(self):
        """
        Pub/sub proxy lazy property with context manager.
        """
        with self._pubsub_lock:
            # Lazy initialization for the pub/sub object.
            if self._pubsub is None:
                self._pubsub = get_redis().pubsub()

        # Yield the pub/sub object (context manager pattern).
        yield self._pubsub

        with self._pubsub_lock:
            # Close and free the pub/sub object if it has no active subscriptions.
            if self._pubsub is not None and len(self._subscriptions) == 0:
                self._pubsub_close()

    @staticmethod
    def _serialize(data: MessageType) -> bytes:
        """
        Serialize a message as bytes before delivering it to either a Redis or websocket channel.
        """
        if isinstance(data, AppMessage):
            data = str(data)
        if isinstance(data, (list, tuple, set)):
            data = list(data)
        if isinstance(data, (list, dict)):
            data = json.dumps(data, cls=AppMessage.Encoder)
        if isinstance(data, str):
            data = data.encode('utf-8')

        return data

    @classmethod
    def publish(cls, data: MessageType, *channels: str) -> None:
        """
        Publish data on one or more Redis channels.
        """
        for channel in channels:
            get_redis().publish(channel, cls._serialize(data))

    def subscribe(self, *channels: str) -> None:
        """
        Subscribe to a set of Redis channels.
        """
        with self.pubsub as pubsub:
            for channel in channels:
                pubsub.subscribe(channel)
                self._subscriptions.add(channel)

    def unsubscribe(self, *channels: str) -> None:
        """
        Unsubscribe from a set of Redis channels.
        """
        with self.pubsub as pubsub:
            for channel in channels:
                if channel in self._subscriptions:
                    pubsub.unsubscribe(channel)
                    self._subscriptions.remove(channel)

    def listen(self) -> Generator[Message, None, None]:
        """
        Listens for pub/sub messages and yields them.
        """
        try:
            with self.pubsub as pubsub:
                pubsub.subscribe(*self._subscriptions)
                for msg in pubsub.listen():
                    channel = msg.get('channel', b'').decode()
                    if msg.get('type') != 'message' or not (
                        channel and channel in self._subscriptions
                    ):
                        continue

                    yield Message(data=msg.get('data', b''), channel=channel)
        except (AttributeError, ValueError, RedisConnectionError):
            return

    def _pubsub_close(self):
        """
        Closes the pub/sub object.
        """
        with self._pubsub_lock:
            if self._pubsub is not None:
                try:
                    self._pubsub.close()
                except Exception as e:
                    logger.debug('Error on pubsub close: %s', e)
                finally:
                    self._pubsub = None

    def on_close(self):
        """
        Extensible close handler that closes the pub/sub object.
        """
        self._pubsub_close()

    @staticmethod
    def get_channel(channel: str) -> str:
        """
        Utility method that returns the prefixed Redis channel for a certain subscription name.
        """
        return f'_platypush/{Config.get("device_id")}/{channel}'  # type: ignore


# vim:sw=4:ts=4:et:



import copy
import json
import importlib
import logging
import threading
from functools import wraps
from typing import Optional, Type

from platypush.common import exec_wrapper
from platypush.config import Config
from platypush.message.event import Event
from platypush.message.request import Request
from platypush.procedure import Procedure
from platypush.utils import get_event_class_by_type, is_functional_hook

logger = logging.getLogger('platypush')


def parse(msg):
    """
    Builds a dict given another dictionary or a JSON UTF-8 encoded
    string/bytearray.
    """

    if isinstance(msg, (bytes, bytearray)):
        msg = msg.decode('utf-8')
    if isinstance(msg, str):
        try:
            msg = json.loads(msg.strip())
        except json.JSONDecodeError:
            logger.warning('Invalid JSON message: %s', msg)
            return None

    return msg


# pylint: disable=too-few-public-methods
class EventCondition:
    """Event hook condition class"""

    # pylint: disable=redefined-builtin
    def __init__(self, type: Optional[Type[Event]] = None, priority=None, **kwargs):
        """
        Rule constructor.
        Params:
            type   -- Class of the event to be built
            kwargs -- Fields rules as a key-value (e.g. source_button=btn_id
                or recognized_phrase='Your phrase')
        """
        self.type = self._get_event_type(type)
        self.args = {}
        self.parsed_args = {}
        self.priority = priority

        for key, value in kwargs.items():
            self.args[key] = value

    @staticmethod
    def _get_event_type(type: Optional[Type[Event]] = None) -> Type[Event]:
        if not type:
            return Event

        # The package alias `platypush.events` -> `platypush.message.event` is
        # supported
        if type.__module__.startswith('platypush.events'):
            module = importlib.import_module(
                'platypush.message.event' + type.__module__[len('platypush.events') :]
            )

            type = getattr(module, type.__name__)
            assert type, f'Invalid event type: {type}'

        return type

    @classmethod
    def build(cls, rule):
        """
        Builds a rule given either another EventRule, a dictionary or a JSON
        UTF-8 encoded string/bytearray.
        """

        if isinstance(rule, cls):
            return rule

        rule = parse(rule)
        assert isinstance(rule, dict), f'Not a valid rule: {rule}'
        type = get_event_class_by_type(rule.pop('type') if 'type' in rule else 'Event')

        args = {}
        for key, value in rule.items():
            args[key] = value

        return cls(type=type, **args)


class EventAction(Request):
    """
    Event hook action class. It is a special type of runnable request whose
    fields can be configured later depending on the event context.
    """

    def __init__(self, target=None, action=None, **args):
        if target is None:
            target = Config.get('device_id')
        args_copy = dict(copy.deepcopy(args))
        super().__init__(target=target, action=action, **args_copy)

    @classmethod
    def build(cls, msg):
        msg = super().parse(msg)
        msg['origin'] = Config.get('device_id')

        if 'target' not in msg:
            msg['target'] = msg['origin']

        token = Config.get('token')
        if token:
            msg['token'] = token

        return super().build(msg)


class EventHook:
    """
    Event hook class. It consists of one conditions and one or multiple actions
    to be executed.
    """

    def __init__(self, name, priority=None, condition=None, actions=None):
        """
        Takes a name, a EventCondition object and an event action procedure as
        input. It may also have a priority attached as a positive number. If
        multiple hooks match against an event, only the ones that have either
        the maximum match score or the maximum pre-configured priority will be
        run.
        """

        self.name = name
        self.condition = EventCondition.build(condition or {})
        self.actions = actions or []
        self.priority = priority or 0
        self.condition.priority = self.priority

    @classmethod
    def build(cls, name, hook):  # pylint: disable=redefined-outer-name
        """
        Builds a rule given either another EventRule, a dictionary or a JSON
        UTF-8 encoded string/bytearray.
        """

        if isinstance(hook, cls):
            return hook

        hook = parse(hook)
        if is_functional_hook(hook):
            actions = Procedure(name=name, requests=[hook], _async=False)
            return cls(
                name=name, condition=getattr(hook, 'condition', None), actions=actions
            )

        assert isinstance(hook, dict)
        condition = EventCondition.build(hook['if']) if 'if' in hook else None
        actions = []
        priority = hook['priority'] if 'priority' in hook else None
        if condition:
            condition.priority = priority

        if 'then' in hook:
            if isinstance(hook['then'], list):
                actions = hook['then']
            else:
                actions = [hook['then']]

        actions = Procedure.build(name=name + '__Hook', requests=actions, _async=False)
        return cls(name=name, condition=condition, actions=actions, priority=priority)

    def matches_event(self, event):
        """
        Returns an EventMatchResult object containing the information about the
        match between the event and this hook.
        """

        return event.matches_condition(self.condition)

    def run(self, event):
        """
        Checks the condition of the hook against a particular event and runs
        the hook actions if the condition is met.
        """

        def _thread_func(result):
            executor = getattr(self.actions, 'execute', None)
            if executor and callable(executor):
                executor(event=event, **result.parsed_args)

        result = self.matches_event(event)

        if result.is_match:
            logger.info(
                'Running hook `%s` triggered by a `%s` event',
                self.name,
                f'{event.__module__}.{event.__class__.__name__}',
            )

            threading.Thread(
                target=_thread_func,
                name='Event-' + self.name,
                args=(result,),
                daemon=True,
            ).start()


def hook(event_type=Event, **condition):
    """
    Decorator used for event hook functions.
    """

    def wrapper(f):
        f.hook = True
        f.condition = EventCondition(type=event_type, **condition)

        @wraps(f)
        def wrapped(event, *args, **kwargs):
            from platypush.message.event.http.hook import WebhookEvent

            response = exec_wrapper(f, event, *args, **kwargs)
            if isinstance(event, WebhookEvent):
                event.send_response(response)

            return response

        return wrapped

    return wrapper


# vim:sw=4:ts=4:et:



import enum
import os
from typing import Optional, Union, Tuple, List

from PIL import Image, ImageFont

from platypush.plugins import Plugin, action


class DeviceInterface(enum.Enum):
    I2C = 'i2c'
    SPI = 'spi'


class DeviceSlot(enum.IntEnum):
    BACK = 0
    FRONT = 1


class DeviceRotation(enum.IntEnum):
    ROTATE_0 = 0
    ROTATE_90 = 1
    ROTATE_180 = 2
    ROTATE_270 = 3


class LumaOledPlugin(Plugin):
    """
    Plugin to interact with small OLED-based RaspberryPi displays through the luma.oled driver.
    """

    def __init__(
        self,
        interface: str,
        device: str,
        port: int = 0,
        slot: int = DeviceSlot.BACK.value,
        width: int = 128,
        height: int = 64,
        rotate: int = DeviceRotation.ROTATE_0.value,
        gpio_DC: int = 24,
        gpio_RST: int = 25,
        bus_speed_hz: int = 8000000,
        address: int = 0x3C,
        cs_high: bool = False,
        transfer_size: int = 4096,
        spi_mode: Optional[int] = None,
        font: Optional[str] = None,
        font_size: int = 10,
        **kwargs
    ):
        """
        :param interface: Serial interface the display is connected to (``spi``
            or ``i2c``).
        :param device: Display chipset type (supported: ssd1306 ssd1309,
            ssd1322, ssd1325, ssd1327, ssd1331, ssd1351, ssd1362, sh1106).
        :param port: Device port (usually 0 or 1).
        :param slot: Device slot (0 for back, 1 for front).
        :param width: Display width.
        :param height: Display height.
        :param rotate: Display rotation (0 for no rotation, 1 for 90 degrees, 2
            for 180 degrees, 3 for 270 degrees).
        :param gpio_DC: [SPI only] GPIO PIN used for data (default: 24).
        :param gpio_RST: [SPI only] GPIO PIN used for RST (default: 25).
        :param bus_speed_hz: [SPI only] Bus speed in Hz (default: 8 MHz).
        :param address: [I2C only] Device address (default: 0x3c).
        :param cs_high: [SPI only] Set to True if the SPI chip select is high.
        :param transfer_size: [SPI only] Maximum amount of bytes to transfer in
            one go (default: 4096).
        :param spi_mode: [SPI only] SPI mode as two bit pattern of clock
            polarity and phase [CPOL|CPHA], 0-3 (default:None).
        :param font: Path to a default TTF font used to display the text.
        :param font_size: Font size - it only applies if ``font`` is set.
        """
        import luma.core.interface.serial as serial
        import luma.oled.device
        from luma.core.render import canvas

        super().__init__(**kwargs)

        iface_name = interface
        interface = getattr(serial, DeviceInterface(interface).value)

        if iface_name == DeviceInterface.SPI.value:
            self.serial = interface(
                port=port,
                device=slot,
                cs_high=cs_high,
                gpio_DC=gpio_DC,
                gpio_RST=gpio_RST,
                bus_speed_hz=bus_speed_hz,
                transfer_size=transfer_size,
                spi_mode=spi_mode,
            )
        else:
            self.serial = interface(port=port, address=address)

        device = getattr(luma.oled.device, device)
        self.device = device(self.serial, width=width, height=height, rotate=rotate)
        self.canvas = canvas(self.device)
        self.font = None
        self.font_size = font_size
        self.font = self._get_font(font, font_size)

    def _get_font(self, font: Optional[str] = None, font_size: Optional[int] = None):
        if font:
            return ImageFont.truetype(
                os.path.abspath(os.path.expanduser(font)), font_size or self.font_size
            )

        return self.font

    @action
    def clear(self):
        """
        clear the display canvas.
        """
        from luma.core.render import canvas

        self.device.clear()
        del self.canvas
        self.canvas = canvas(self.device)

    @action
    def text(
        self,
        text: str,
        pos: Union[Tuple[int], List[int]] = (0, 0),
        fill: str = 'white',
        font: Optional[str] = None,
        font_size: Optional[int] = None,
        clear: bool = False,
    ):
        """
        Draw text on the canvas.

        :param text: Text to be drawn.
        :param pos: Position of the text.
        :param fill: Text color (default: ``white``).
        :param font: ``font`` type override.
        :param font_size: ``font_size`` override.
        :param clear: Set to True if you want to clear the canvas before
            writing the text (default: False).
        """
        if clear:
            self.clear()

        font = self._get_font(font, font_size)

        with self.canvas as draw:
            draw.text(pos, text, fill=fill, font=font)

    @action
    def rectangle(
        self,
        xy: Optional[Union[Tuple[int], List[int]]] = None,
        fill: Optional[str] = None,
        outline: Optional[str] = None,
        width: int = 1,
        clear: bool = False,
    ):
        """
        Draw a rectangle on the canvas.

        :param xy: Two points defining the bounding box, either as ``[(x0, y0),
            (x1, y1)]`` or ``[x0, y0, x1, y1]``. Default: bounding box of the
            device.
        :param fill: Fill color - can be ``black`` or ``white``.
        :param outline: Outline color - can be ``black`` or ``white``.
        :param width: Figure width in pixels (default: 1).
        :param clear: Set to True if you want to clear the canvas before
            writing the text (default: False).
        """
        if clear:
            self.clear()

        if not xy:
            xy = self.device.bounding_box

        with self.canvas as draw:
            draw.rectangle(xy, outline=outline, fill=fill, width=width)

    @action
    def arc(
        self,
        start: int,
        end: int,
        xy: Optional[Union[Tuple[int], List[int]]] = None,
        fill: Optional[str] = None,
        outline: Optional[str] = None,
        width: int = 1,
        clear: bool = False,
    ):
        """
        Draw an arc on the canvas.

        :param start: Starting angle, in degrees (measured from 3 o' clock and
            increasing clockwise).
        :param end: Ending angle, in degrees (measured from 3 o' clock and
            increasing clockwise).
        :param xy: Two points defining the bounding box, either as ``[(x0, y0),
            (x1, y1)]`` or ``[x0, y0, x1, y1]``. Default: bounding box of the
            device.
        :param fill: Fill color - can be ``black`` or ``white``.
        :param outline: Outline color - can be ``black`` or ``white``.
        :param width: Figure width in pixels (default: 1).
        :param clear: Set to True if you want to clear the canvas before
            writing the text (default: False).
        """
        if clear:
            self.clear()

        if not xy:
            xy = self.device.bounding_box

        with self.canvas as draw:
            draw.arc(xy, start=start, end=end, outline=outline, fill=fill, width=width)

    @action
    def chord(
        self,
        start: int,
        end: int,
        xy: Optional[Union[Tuple[int], List[int]]] = None,
        fill: Optional[str] = None,
        outline: Optional[str] = None,
        width: int = 1,
        clear: bool = False,
    ):
        """
        Same as ``arc``, but it connects the end points with a straight line.

        :param start: Starting angle, in degrees (measured from 3 o' clock and
            increasing clockwise).
        :param end: Ending angle, in degrees (measured from 3 o' clock and
            increasing clockwise).
        :param xy: Two points defining the bounding box, either as ``[(x0, y0),
            (x1, y1)]`` or ``[x0, y0, x1, y1]``. Default: bounding box of the
            device.
        :param fill: Fill color - can be ``black`` or ``white``.
        :param outline: Outline color - can be ``black`` or ``white``.
        :param width: Figure width in pixels (default: 1).
        :param clear: Set to True if you want to clear the canvas before
            writing the text (default: False).
        """
        if clear:
            self.clear()

        if not xy:
            xy = self.device.bounding_box

        with self.canvas as draw:
            draw.chord(
                xy, start=start, end=end, outline=outline, fill=fill, width=width
            )

    @action
    def pieslice(
        self,
        start: int,
        end: int,
        xy: Optional[Union[Tuple[int], List[int]]] = None,
        fill: Optional[str] = None,
        outline: Optional[str] = None,
        width: int = 1,
        clear: bool = False,
    ):
        """
        Same as ``arc``, but it also draws straight lines between the end
        points and the center of the bounding box.

        :param start: Starting angle, in degrees (measured from 3 o' clock and
            increasing clockwise).
        :param end: Ending angle, in degrees (measured from 3 o' clock and
            increasing clockwise).
        :param xy: Two points defining the bounding box, either as ``[(x0, y0),
            (x1, y1)]`` or ``[x0, y0, x1, y1]``. Default: bounding box of the
            device.
        :param fill: Fill color - can be ``black`` or ``white``.
        :param outline: Outline color - can be ``black`` or ``white``.
        :param width: Figure width in pixels (default: 1).
        :param clear: Set to True if you want to clear the canvas before
            writing the text (default: False).
        """
        if clear:
            self.clear()

        if not xy:
            xy = self.device.bounding_box

        with self.canvas as draw:
            draw.pieslice(
                xy, start=start, end=end, outline=outline, fill=fill, width=width
            )

    @action
    def ellipse(
        self,
        xy: Optional[Union[Tuple[int], List[int]]] = None,
        fill: Optional[str] = None,
        outline: Optional[str] = None,
        width: int = 1,
        clear: bool = False,
    ):
        """
        Draw an ellipse on the canvas.

        :param xy: Two points defining the bounding box, either as ``[(x0, y0),
            (x1, y1)]`` or ``[x0, y0, x1, y1]``. Default: bounding box of the
            device.
        :param fill: Fill color - can be ``black`` or ``white``.
        :param outline: Outline color - can be ``black`` or ``white``.
        :param width: Figure width in pixels (default: 1).
        :param clear: Set to True if you want to clear the canvas before
            writing the text (default: False).
        """
        if clear:
            self.clear()

        if not xy:
            xy = self.device.bounding_box

        with self.canvas as draw:
            draw.ellipse(xy, outline=outline, fill=fill, width=width)

    @action
    def line(
        self,
        xy: Optional[Union[Tuple[int], List[int]]] = None,
        fill: Optional[str] = None,
        outline: Optional[str] = None,
        width: int = 1,
        curve: bool = False,
        clear: bool = False,
    ):
        """
        Draw a line on the canvas.

        :param xy: Two points defining the bounding box, either as ``[(x0, y0),
            (x1, y1)]`` or ``[x0, y0, x1, y1]``. Default: bounding box of the
            device.
        :param fill: Fill color - can be ``black`` or ``white``.
        :param outline: Outline color - can be ``black`` or ``white``.
        :param width: Figure width in pixels (default: 1).
        :param curve: Set to True for rounded edges (default: False).
        :param clear: Set to True if you want to clear the canvas before
            writing the text (default: False).
        """
        if clear:
            self.clear()

        if not xy:
            xy = self.device.bounding_box

        with self.canvas as draw:
            draw.line(
                xy,
                outline=outline,
                fill=fill,
                width=width,
                joint='curve' if curve else None,
            )

    @action
    def point(
        self,
        xy: Optional[Union[Tuple[int], List[int]]] = None,
        fill: Optional[str] = None,
        clear: bool = False,
    ):
        """
        Draw one or more points on the canvas.

        :param xy: Two points defining the bounding box, either as ``[(x0, y0),
            (x1, y1)]`` or ``[x0, y0, x1, y1]``. Default: bounding box of the
            device.
        :param fill: Fill color - can be ``black`` or ``white``.
        :param clear: Set to True if you want to clear the canvas before
            writing the text (default: False).
        """
        if clear:
            self.clear()

        if not xy:
            xy = self.device.bounding_box

        with self.canvas as draw:
            draw.point(xy, fill=fill)

    @action
    def polygon(
        self,
        xy: Optional[Union[Tuple[int], List[int]]] = None,
        fill: Optional[str] = None,
        outline: Optional[str] = None,
        clear: bool = False,
    ):
        """
        Draw a polygon on the canvas.

        :param xy: Two points defining the bounding box, either as ``[(x0, y0),
            (x1, y1)]`` or ``[x0, y0, x1, y1]``. Default: bounding box of the
            device.
        :param fill: Fill color - can be ``black`` or ``white``.
        :param outline: Outline color - can be ``black`` or ``white``.
        :param clear: Set to True if you want to clear the canvas before
            writing the text (default: False).
        """
        if clear:
            self.clear()

        if not xy:
            xy = self.device.bounding_box

        with self.canvas as draw:
            draw.polygon(xy, outline=outline, fill=fill)

    @action
    def image(self, image: str):
        """
        Draws an image to the canvas (this will clear the existing canvas).

        :param image: Image path.
        """
        image = Image.open(os.path.abspath(os.path.expanduser(image)))
        self.clear()
        self.device.display(image)


# vim:sw=4:ts=4:et:



from dataclasses import dataclass, field


@dataclass
class Fan:
    """
    System fan sensor data class.
    """

    id: str = field(
        metadata={
            'metadata': {
                'description': 'Unique ID for the sensor',
                'example': 'acpi_1',
            }
        }
    )

    label: str = field(
        metadata={
            'metadata': {
                'description': 'Name of the sensor',
                'example': 'CPU',
            }
        }
    )

    value: float = field(
        metadata={
            'metadata': {
                'description': 'Current fan speed, in RPM',
                'example': 3000,
            }
        }
    )



import json
import os
import pathlib

from datetime import datetime
from typing import Iterable, Optional, Union

from platypush.config import Config
from platypush.context import Variable, get_bus
from platypush.message.event.music.tidal import TidalPlaylistUpdatedEvent
from platypush.plugins import RunnablePlugin, action
from platypush.plugins.music.tidal.workers import get_items
from platypush.schemas.tidal import (
    TidalAlbumSchema,
    TidalPlaylistSchema,
    TidalArtistSchema,
    TidalSearchResultsSchema,
    TidalTrackSchema,
)


class MusicTidalPlugin(RunnablePlugin):
    """
    Plugin to interact with the user's Tidal account and library.

    Upon the first login, the application will prompt you with a link to
    connect to your Tidal account. Once authorized, you should no longer be
    required to explicitly login.
    """

    _base_url = 'https://api.tidalhifi.com/v1/'
    _default_credentials_file = os.path.join(
        str(Config.get('workdir')), 'tidal', 'credentials.json'
    )

    def __init__(
        self,
        quality: str = 'high',
        credentials_file: str = _default_credentials_file,
        **kwargs,
    ):
        """
        :param quality: Default audio quality. Default: ``high``.
            Supported: [``loseless``, ``master``, ``high``, ``low``].
        :param credentials_file: Path to the file where the OAuth session
            parameters will be stored (default:
            ``<WORKDIR>/tidal/credentials.json``).
        """

        super().__init__(**kwargs)
        self._credentials_file = os.path.expanduser(credentials_file)
        self._user_playlists = {}
        self._quality = self._get_quality(quality)
        self._session = None

    @staticmethod
    def _get_quality(quality: str):
        from tidalapi import Quality

        try:
            return getattr(Quality, quality.lower())
        except AttributeError:
            try:
                return Quality(quality.upper())
            except ValueError as e:
                raise AssertionError(
                    f'Invalid quality: {quality}. Supported values: '
                    f'{[q.value for q in Quality]}'
                ) from e

    def _oauth_open_saved_session(self):
        if not self._session:
            return

        try:
            with open(self._credentials_file, 'r') as f:
                data = json.load(f)
                self._session.load_oauth_session(
                    data['token_type'], data['access_token'], data['refresh_token']
                )
        except Exception as e:
            self.logger.warning('Could not load %s: %s', self._credentials_file, e)

    def _oauth_create_new_session(self):
        if not self._session:
            return

        self._session.login_oauth_simple(function=self.logger.warning)  # type: ignore
        if self._session.check_login():
            data = {
                'token_type': self._session.token_type,
                'session_id': self._session.session_id,
                'access_token': self._session.access_token,
                'refresh_token': self._session.refresh_token,
            }

            pathlib.Path(os.path.dirname(self._credentials_file)).mkdir(
                parents=True, exist_ok=True
            )

            with open(self._credentials_file, 'w') as outfile:
                json.dump(data, outfile)

    @staticmethod
    def _ensure_user_playlist(playlist):
        from tidalapi import UserPlaylist

        assert isinstance(
            playlist, UserPlaylist
        ), 'This operation is only possible on user playlists'
        return playlist

    @property
    def session(self):
        from tidalapi import Config, Session

        if self._session and self._session.check_login():
            return self._session

        # Attempt to reload the existing session from file
        self._session = Session(config=Config(quality=self._quality))
        self._oauth_open_saved_session()
        if not self._session.check_login():
            # Create a new session if we couldn't load an existing one
            self._oauth_create_new_session()

        assert (
            self._session.user and self._session.check_login()
        ), 'Could not connect to TIDAL'

        return self._session

    @property
    def user(self):
        user = self.session.user
        assert user, 'Not logged in'
        return user

    @action
    def create_playlist(self, name: str, description: Optional[str] = None):
        """
        Create a new playlist.

        :param name: Playlist name.
        :param description: Optional playlist description.
        :return: .. schema:: tidal.TidalPlaylistSchema
        """
        ret = self.user.create_playlist(name, description)
        return TidalPlaylistSchema().dump(ret)

    @action
    def delete_playlist(self, playlist_id: str):
        """
        Delete a playlist by ID.

        :param playlist_id: ID of the playlist to delete.
        """
        pl = self.session.playlist(playlist_id)
        pl = self._ensure_user_playlist(pl)
        pl.delete()

    @action
    def edit_playlist(self, playlist_id: str, title=None, description=None):
        """
        Edit a playlist's metadata.

        :param name: New name.
        :param description: New description.
        """
        pl = self.session.playlist(playlist_id)
        pl = self._ensure_user_playlist(pl)
        pl.edit(title=title, description=description)

    @action
    def get_playlists(self):
        """
        Get the user's playlists (track lists are excluded).

        :return: .. schema:: tidal.TidalPlaylistSchema(many=True)
        """
        ret = self.user.playlists() + self.user.favorites.playlists()
        return TidalPlaylistSchema().dump(ret, many=True)

    @action
    def get_playlist(self, playlist_id: str):
        """
        Get the details of a playlist (including tracks).

        :param playlist_id: Playlist ID.
        :return: .. schema:: tidal.TidalPlaylistSchema
        """
        pl = self.session.playlist(playlist_id)
        pl._tracks = get_items(pl.tracks)  # type: ignore
        return TidalPlaylistSchema().dump(pl)

    @action
    def get_artist(self, artist_id: Union[str, int]):
        """
        Get the details of an artist.

        :param artist_id: Artist ID.
        :return: .. schema:: tidal.TidalArtistSchema
        """
        ret = self.session.artist(artist_id)
        ret.albums = get_items(ret.get_albums)  # type: ignore
        return TidalArtistSchema().dump(ret)

    @action
    def get_album(self, album_id: Union[str, int]):
        """
        Get the details of an album.

        :param artist_id: Album ID.
        :return: .. schema:: tidal.TidalAlbumSchema
        """
        ret = self.session.album(album_id)
        return TidalAlbumSchema(with_tracks=True).dump(ret)

    @action
    def get_track(self, track_id: Union[str, int]):
        """
        Get the details of an track.

        :param artist_id: Track ID.
        :return: .. schema:: tidal.TidalTrackSchema
        """
        ret = self.session.album(track_id)
        return TidalTrackSchema().dump(ret)

    @action
    def search(
        self,
        query: str,
        limit: int = 50,
        offset: int = 0,
        type: Optional[str] = None,
    ):
        """
        Perform a search.

        :param query: Query string.
        :param limit: Maximum results that should be returned (default: 50).
        :param offset: Search offset (default: 0).
        :param type: Type of results that should be returned. Default: None
            (return all the results that match the query). Supported:
            ``artist``, ``album``, ``track`` and ``playlist``.
        :return: .. schema:: tidal.TidalSearchResultsSchema
        """
        from tidalapi.artist import Artist
        from tidalapi.album import Album
        from tidalapi.media import Track
        from tidalapi.playlist import Playlist

        models = None
        if type is not None:
            if type == 'artist':
                models = [Artist]
            elif type == 'album':
                models = [Album]
            elif type == 'track':
                models = [Track]
            elif type == 'playlist':
                models = [Playlist]
            else:
                raise AssertionError(f'Unsupported search type: {type}')

        ret = self.session.search(query, models=models, limit=limit, offset=offset)

        return TidalSearchResultsSchema().dump(ret)

    @action
    def get_download_url(self, track_id: str) -> str:
        """
        Get the direct download URL of a track.

        :param artist_id: Track ID.
        """
        return self.session.track(track_id).get_url()

    @action
    def add_to_playlist(self, playlist_id: str, track_ids: Iterable[Union[str, int]]):
        """
        Append one or more tracks to a playlist.

        :param playlist_id: Target playlist ID.
        :param track_ids: List of track IDs to append.
        """
        pl = self.session.playlist(playlist_id)
        pl = self._ensure_user_playlist(pl)
        pl.add(track_ids)

    @action
    def remove_from_playlist(
        self,
        playlist_id: str,
        track_id: Optional[Union[str, int]] = None,
        index: Optional[int] = None,
    ):
        """
        Remove a track from a playlist.

        Specify either the ``track_id`` or the ``index``.

        :param playlist_id: Target playlist ID.
        :param track_id: ID of the track to remove.
        :param index: Index of the track to remove.
        """
        assert not (
            track_id is None and index is None
        ), 'Please specify either track_id or index'

        pl = self.session.playlist(playlist_id)
        pl = self._ensure_user_playlist(pl)

        if index:
            pl.remove_by_index(index)
        if track_id:
            pl.remove_by_id(track_id)

    @action
    def add_track(self, track_id: Union[str, int]):
        """
        Add a track to the user's collection.

        :param track_id: Track ID.
        """
        self.user.favorites.add_track(track_id)

    @action
    def add_album(self, album_id: Union[str, int]):
        """
        Add an album to the user's collection.

        :param album_id: Album ID.
        """
        self.user.favorites.add_album(album_id)

    @action
    def add_artist(self, artist_id: Union[str, int]):
        """
        Add an artist to the user's collection.

        :param artist_id: Artist ID.
        """
        self.user.favorites.add_artist(artist_id)

    @action
    def add_playlist(self, playlist_id: str):
        """
        Add a playlist to the user's collection.

        :param playlist_id: Playlist ID.
        """
        self.user.favorites.add_playlist(playlist_id)

    @action
    def remove_track(self, track_id: Union[str, int]):
        """
        Remove a track from the user's collection.

        :param track_id: Track ID.
        """
        self.user.favorites.remove_track(track_id)

    @action
    def remove_album(self, album_id: Union[str, int]):
        """
        Remove an album from the user's collection.

        :param album_id: Album ID.
        """
        self.user.favorites.remove_album(album_id)

    @action
    def remove_artist(self, artist_id: Union[str, int]):
        """
        Remove an artist from the user's collection.

        :param artist_id: Artist ID.
        """
        self.user.favorites.remove_artist(artist_id)

    @action
    def remove_playlist(self, playlist_id: str):
        """
        Remove a playlist from the user's collection.

        :param playlist_id: Playlist ID.
        """
        self.user.favorites.remove_playlist(playlist_id)

    def main(self):
        while not self.should_stop():
            playlists = self.session.user.playlists()  # type: ignore

            for pl in playlists:
                last_updated_var = Variable(f'TIDAL_PLAYLIST_LAST_UPDATE[{pl.id}]')
                prev_last_updated = last_updated_var.get()
                if prev_last_updated:
                    prev_last_updated = datetime.fromisoformat(prev_last_updated)
                    if pl.last_updated > prev_last_updated:
                        get_bus().post(TidalPlaylistUpdatedEvent(playlist_id=pl.id))

                if not prev_last_updated or pl.last_updated > prev_last_updated:
                    last_updated_var.set(pl.last_updated.isoformat())

            self.wait_stop(self.poll_interval)



from collections import defaultdict
from logging import getLogger
from threading import RLock
from typing import Dict, Iterable, List, Optional, Union

from .._model import AudioDevice, DeviceType, StreamType
from .._streams import AudioThread
from ._device import DeviceManager


class StreamManager:
    """
    The audio manager is responsible for storing the current state of the
    playing/recording audio streams and allowing fast flexible lookups (by
    stream index, name, type, device, and any combination of those).
    """

    def __init__(self, device_manager: DeviceManager):
        """
        :param device_manager: Reference to the device manager.
        """
        self._next_stream_index = 1
        self._device_manager = device_manager
        self._state_lock = RLock()
        self._stream_index_by_name: Dict[str, int] = {}
        self._stream_name_by_index: Dict[int, str] = {}
        self._stream_index_to_device: Dict[int, AudioDevice] = {}
        self._stream_index_to_type: Dict[int, StreamType] = {}
        self.logger = getLogger(__name__)

        self._streams: Dict[
            int, Dict[StreamType, Dict[int, AudioThread]]
        ] = defaultdict(lambda: {stream_type: {} for stream_type in StreamType})
        """ {device_index: {stream_type: {stream_index: audio_thread}}} """

        self._streams_by_index: Dict[StreamType, Dict[int, AudioThread]] = {
            stream_type: {} for stream_type in StreamType
        }
        """ {stream_type: {stream_index: [audio_threads]}} """

        self._stream_locks: Dict[int, Dict[StreamType, RLock]] = defaultdict(
            lambda: {stream_type: RLock() for stream_type in StreamType}
        )
        """ {device_index: {stream_type: RLock}} """

    @classmethod
    def _generate_stream_name(
        cls,
        type: StreamType,  # pylint: disable=redefined-builtin
        stream_index: int,
    ) -> str:
        return f'platypush:audio:{type.value}:{stream_index}'

    def _gen_next_stream_index(
        self,
        type: StreamType,  # pylint: disable=redefined-builtin
        stream_name: Optional[str] = None,
    ) -> int:
        """
        :param type: The type of the stream to allocate (input or output).
        :param stream_name: The name of the stream to allocate.
        :return: The index of the new stream.
        """
        with self._state_lock:
            stream_index = self._next_stream_index

            if not stream_name:
                stream_name = self._generate_stream_name(type, stream_index)

            self._stream_name_by_index[stream_index] = stream_name
            self._stream_index_by_name[stream_name] = stream_index
            self._next_stream_index += 1

        return stream_index

    def register(
        self,
        audio_thread: AudioThread,
        device: AudioDevice,
        type: StreamType,  # pylint: disable=redefined-builtin
        stream_name: Optional[str] = None,
    ):
        """
        Registers an audio stream to a device.

        :param audio_thread: Stream to register.
        :param device: Device to register the stream to.
        :param type: The type of the stream to allocate (input or output).
        :param stream_name: The name of the stream to allocate.
        """
        with self._state_lock:
            stream_index = audio_thread.stream_index
            if stream_index is None:
                stream_index = audio_thread.stream_index = self._gen_next_stream_index(
                    type, stream_name=stream_name
                )

            self._streams[device.index][type][stream_index] = audio_thread
            self._stream_index_to_device[stream_index] = device
            self._stream_index_to_type[stream_index] = type
            self._streams_by_index[type][stream_index] = audio_thread

    def unregister(
        self,
        audio_thread: AudioThread,
        device: Optional[AudioDevice] = None,
        type: Optional[StreamType] = None,  # pylint: disable=redefined-builtin
    ):
        """
        Unregisters an audio stream from a device.

        :param audio_thread: Stream to unregister.
        :param device: Device to unregister the stream from.
        :param type: The type of the stream to unregister (input or output).
        """
        with self._state_lock:
            stream_index = audio_thread.stream_index
            if stream_index is None:
                return

            if device is None:
                device = self._stream_index_to_device.get(stream_index)

            if not type:
                type = self._stream_index_to_type.get(stream_index)

            if device is None or type is None:
                return

            self._streams[device.index][type].pop(stream_index, None)
            self._stream_index_to_device.pop(stream_index, None)
            self._stream_index_to_type.pop(stream_index, None)
            self._streams_by_index[type].pop(stream_index, None)
            stream_name = self._stream_name_by_index.pop(stream_index, None)
            if stream_name:
                self._stream_index_by_name.pop(stream_name, None)

    def _get_by_device_and_type(
        self,
        device: Optional[DeviceType] = None,
        type: Optional[StreamType] = None,  # pylint: disable=redefined-builtin
    ) -> List[AudioThread]:
        """
        Filter streams by device and/or type.
        """
        devs = (
            [self._device_manager.get_device(device, type)]
            if device is not None
            else self._device_manager.get_devices(type)
        )

        return [
            audio_thread
            for dev in devs
            for stream_info in (
                [self._streams[dev.index].get(type, {})]
                if type
                else list(self._streams[dev.index].values())
            )
            for audio_thread in stream_info.values()
            if audio_thread and audio_thread.is_alive()
        ]

    def _get_by_stream_index_or_name(
        self, streams: Iterable[Union[str, int]]
    ) -> List[AudioThread]:
        """
        Filter streams by index or name.
        """
        threads = []

        for stream in streams:
            try:
                stream_index = int(stream)
            except (TypeError, ValueError):
                stream_index = self._stream_index_by_name.get(stream)  # type: ignore
                if stream_index is None:
                    self.logger.warning('No such audio stream: %s', stream)
                    continue

            stream_type = self._stream_index_to_type.get(stream_index)
            if not stream_type:
                self.logger.warning(
                    'No type available for this audio stream: %s', stream
                )
                continue

            thread = self._streams_by_index.get(stream_type, {}).get(stream_index)
            if thread:
                threads.append(thread)

        return threads

    def get(
        self,
        device: Optional[DeviceType] = None,
        type: Optional[StreamType] = None,  # pylint: disable=redefined-builtin
        streams: Optional[Iterable[Union[str, int]]] = None,
    ) -> List[AudioThread]:
        """
        Searches streams, either by device and/or type, or by stream index/name.
        """
        return (
            self._get_by_stream_index_or_name(streams)
            if streams
            else self._get_by_device_and_type(device, type)
        )



from base64 import b64encode
import json
import subprocess
import threading
import time
from typing import Callable, Optional
from uuid import uuid1

from platypush.common.cmd_stream import redis_topic
from platypush.plugins import Plugin, action
from platypush.utils import get_redis


class ShellPlugin(Plugin):
    """
    Plugin to run shell commands.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _exec(self, func: Callable, cmd: str, ignore_errors: bool = False):
        try:
            return func(cmd)
        except subprocess.CalledProcessError as e:
            if ignore_errors:
                self.logger.warning(
                    'Command %s failed with error: %s', cmd, e.output.decode()
                )
                return None

            raise RuntimeError(e.output.decode()) from e

    def _exec_simple(self, cmd: str):
        return subprocess.check_output(
            cmd, stderr=subprocess.STDOUT, shell=True
        ).decode()

    @staticmethod
    def _send_ws_output(cmd_id: str, buf: Optional[bytes]):
        get_redis().publish(
            redis_topic,
            json.dumps(
                {
                    'id': cmd_id,
                    'output': (b64encode(buf).decode() if buf is not None else None),
                }
            ),
        )

    def _exec_ws_thread(self, cmd: str, cmd_id: str):
        try:
            with subprocess.Popen(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
            ) as proc:
                first_loop = True

                while proc.poll() is None and proc.stdout and proc.stdout.readable():
                    buf = b''
                    bufsize = 1024

                    while len(buf) < bufsize:
                        ch = proc.stdout.read(1)
                        if not ch:
                            break

                        buf += ch
                        if ch in (b'\r', b'\n'):
                            break

                    if first_loop:
                        first_loop = False
                        # An artificial delay to let the client connect to the websocket,
                        # otherwise the first few lines of output might be lost.
                        time.sleep(2)

                    self._send_ws_output(cmd_id, buf)
        finally:
            self._send_ws_output(cmd_id, None)

    def _exec_ws(self, cmd: str):
        cmd_id = str(uuid1())
        threading.Thread(target=self._exec_ws_thread, args=(cmd, cmd_id)).start()
        return {'ws_path': f'/ws/shell?id={cmd_id}'}

    @action
    def exec(
        self,
        cmd: str,
        background: bool = False,
        ws: bool = False,
        ignore_errors: bool = False,
    ):
        """
        Run a command.

        :param cmd: Command to execute
        :param background: If set to True, execute the process in the
            background, otherwise wait for the process termination and return
            its output (default: False).
        :param ignore_errors: If set, then any errors in the command execution
            will be ignored. Otherwise a RuntimeError will be thrown (default
            value: False)
        :param ws: If set to True then the output of the command will be
            sent asynchronously over a websocket channel (default: False).
            In this case, the method will return a response in the format:

                .. code-block:: json

                    {
                        "ws_path": "/ws/shell?id=<cmd_id>"
                    }

            Where ``ws_path`` is the websocket path where the output of the
            command will be sent. The websocket will be closed when the command
            terminates.

        :returns: A response object where the ``output`` field will contain the
            command output as a string, and the ``errors`` field will contain
            whatever was sent to stderr.
        """
        if background:
            subprocess.Popen(cmd, shell=True)
            return None

        func = self._exec_ws if ws else self._exec_simple
        return self._exec(func, cmd, ignore_errors=ignore_errors)


# vim:sw=4:ts=4:et:



from typing import Collection, List, Optional

from linode_api4 import LinodeClient, Instance, objects

from platypush.context import get_bus
from platypush.entities.cloud import CloudInstance
from platypush.entities.managers.cloud import CloudInstanceEntityManager, InstanceId
from platypush.entities.managers.switches import EnumSwitchEntityManager
from platypush.entities.switches import EnumSwitch
from platypush.message.event.linode import LinodeInstanceStatusChanged
from platypush.schemas.linode import (
    LinodeInstance,
    LinodeInstanceSchema,
    LinodeInstanceStatus,
)
from platypush.plugins import RunnablePlugin, action


class LinodePlugin(RunnablePlugin, CloudInstanceEntityManager, EnumSwitchEntityManager):
    """
    This plugin can interact with a Linode account and manage node and volumes.

    To get your token:

        - Login to <https://cloud.linode.com/>.
        - Go to My Profile -> API Tokens -> Add a Personal Access Token.
        - Select the scopes that you want to provide to your new token.

    """

    def __init__(self, token: str, poll_interval: float = 60.0, **kwargs):
        """
        :param token: Linode API token.
        :param poll_interval: How often to poll the Linode API
            (default: 60 seconds).
        """
        super().__init__(poll_interval=poll_interval, **kwargs)
        self._token = token

    def _get_client(self, token: Optional[str] = None) -> LinodeClient:
        """
        Get a :class:`LinodeClient` instance.

        :param token: Override the default token.
        """
        return LinodeClient(token or self._token)

    def _get_instance(
        self, instance: InstanceId, token: Optional[str] = None
    ) -> Instance:
        """
        Get an instance by name or ID.

        :param instance: The label, ID or host UUID of the instance.
        :param token: Override the default token.
        """
        client = self._get_client(token)
        if isinstance(instance, str):
            filters = [Instance.label == instance]
        elif isinstance(instance, int):
            filters = [Instance.id == instance]
        else:
            raise AssertionError(f'Invalid instance type: {type(instance)}')

        instances = client.linode.instances(*filters)
        assert instances, f'No such Linode instance: {instance}'
        return instances[0]

    def _linode_instance_to_dict(self, instance: Instance) -> dict:
        """
        Convert an internal :class:`linode_api4.Instance` to a
        dictionary representation that can be used to create a
        :class:`platypush.entities.cloud.CloudInstance` object.
        """
        return {
            key: (value.dict if isinstance(value, objects.MappedObject) else value)
            for key, value in instance.__dict__.items()
            if not key.startswith('_')
        }

    def main(self):
        instances = []

        while not self.should_stop():
            status = {instance.id: instance for instance in instances}

            new_status = {
                instance.id: instance
                for instance in self.status(publish_entities=False).output
            }

            changed_instances = (
                [
                    instance
                    for instance in new_status.values()
                    if not (
                        status.get(instance.id)
                        and status[instance.id].status == instance.status
                    )
                ]
                if new_status
                else []
            )

            if changed_instances:
                for instance in changed_instances:
                    get_bus().post(
                        LinodeInstanceStatusChanged(
                            instance_id=instance.id,
                            instance_name=instance.name,
                            status=instance.status,
                            old_status=(
                                status[instance.id].status
                                if status.get(instance.id)
                                else None
                            ),
                        )
                    )

                self.publish_entities(changed_instances)

            instances = new_status.values()
            self.wait_stop(self.poll_interval)

    def transform_entities(
        self, entities: Collection[LinodeInstance]
    ) -> Collection[CloudInstance]:
        schema = LinodeInstanceSchema()
        return super().transform_entities(
            [
                CloudInstance(
                    reachable=instance.status == LinodeInstanceStatus.RUNNING,
                    children=[
                        EnumSwitch(
                            id=f'{instance.id}:__action__',
                            name='Actions',
                            values=['boot', 'reboot', 'shutdown'],
                            is_write_only=True,
                        )
                    ],
                    **schema.dump(instance),
                )
                for instance in entities
            ]
        )

    @action
    def status(
        self,
        *_,
        instance: Optional[InstanceId] = None,
        token: Optional[str] = None,
        publish_entities: bool = True,
        **__,
    ) -> List[LinodeInstance]:
        """
        Get the full status and info of the instances associated to a selected account.

        :param token: Override the default access token if you want to query another account.
        :param instance: Select only one instance, either by name, ID or host UUID.
        :param publish_entities: Whether
            :class:`platypush.message.event.entities.EntityUpdateEvent` should
            be published for all the instances, whether or not their status has
            changed (default: ``True``).
        :return: .. schema:: linode.LinodeInstanceSchema(many=True)
        """
        instances = (
            [self._get_instance(instance=instance)]
            if instance
            else [
                instance
                for page in self._get_client(token).linode.instances().lists
                for instance in page
            ]
        )

        mapped_instances = LinodeInstanceSchema(many=True).load(
            map(self._linode_instance_to_dict, instances)
        )

        if publish_entities:
            self.publish_entities(mapped_instances)

        return mapped_instances

    @action
    def reboot(self, instance: InstanceId, token: Optional[str] = None, **_):
        """
        Reboot an instance.

        :param instance: Instance ID, label or host UUID.
        :param token: Default access token override.
        """
        node = self._get_instance(instance=instance, token=token)
        assert node.reboot(), 'Reboot failed'

    @action
    def boot(self, instance: InstanceId, token: Optional[str] = None, **_):
        """
        Boot an instance.

        :param instance: Instance ID, label or host UUID.
        :param token: Default access token override.
        """
        node = self._get_instance(instance=instance, token=token)
        assert node.boot(), 'Boot failed'

    @action
    def shutdown(self, instance: InstanceId, token: Optional[str] = None, **_):
        """
        Shutdown an instance.

        :param instance: Instance ID, label or host UUID.
        :param token: Default access token override.
        """
        node = self._get_instance(instance=instance, token=token)
        assert node.shutdown(), 'Shutdown failed'

    @action
    def set(self, entity: str, value: str, **kwargs):
        """
        Entity framework compatible method to run an action on the instance
        through an ``EnumSwitch``.

        :param entity: Entity ID, as ``<instance_id>`` or
            ``<instance_id>:__action__``.
        :param value: Action to run, one among ``boot``, ``reboot`` and
            ``shutdown``.
        """
        try:
            instance_id = int(entity.removesuffix(':__action__'))
        except (TypeError, ValueError) as e:
            raise AssertionError(f'Invalid entity: {entity}') from e

        assert value in {'boot', 'reboot', 'shutdown'}, f'Invalid action: {value}'
        method = getattr(self, value)
        return method(instance_id, **kwargs)


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from typing import List, Optional
from platypush.entities.managers.weather import WeatherEntityManager

from platypush.message.event.weather import (
    NewWeatherConditionEvent,
    NewWeatherForecastEvent,
)
from platypush.plugins import RunnablePlugin, action
from platypush.schemas.weather.openweathermap import WeatherReportSchema
from platypush.utils import get_plugin_name_by_class


class WeatherPlugin(RunnablePlugin, WeatherEntityManager, ABC):
    """
    Base class for weather plugins.
    """

    def __init__(self, poll_interval: Optional[float] = 120, **kwargs):
        super().__init__(poll_interval=poll_interval, **kwargs)
        self._latest_weather = None
        self._latest_forecast = None

    def _on_weather_data(self, weather: dict, always_publish: bool = False):
        if weather != self._latest_weather or always_publish:
            self._bus.post(
                NewWeatherConditionEvent(
                    plugin_name=get_plugin_name_by_class(self.__class__), **weather
                )
            )

            self.publish_entities([weather], type='weather')

        self._latest_weather = weather

    def _on_weather_forecast(self, forecast: List[dict], always_publish: bool = False):
        if forecast != self._latest_forecast or always_publish:
            self._bus.post(
                NewWeatherForecastEvent(
                    plugin_name=get_plugin_name_by_class(self.__class__),
                    forecast=forecast,
                )
            )

            self.publish_entities(forecast, type='forecast')

        self._latest_forecast = forecast

    @action
    def get_current_weather(
        self,
        *args,
        lat: Optional[float] = None,
        long: Optional[float] = None,
        units: Optional[str] = None,
        **kwargs
    ) -> dict:
        """
        Returns the current weather.

        :param lat: Override the ``lat`` configuration value.
        :param long: Override the ``long`` configuration value.
        :param units: Override the ``units`` configuration value.
        :return: .. schema:: weather.openweathermap.WeatherSchema
        """
        weather = self._get_current_weather(
            *args, lat=lat, long=long, units=units, **kwargs
        )
        self._on_weather_data(weather, always_publish=True)
        return weather

    @action
    def get_forecast(
        self,
        *args,
        lat: Optional[float] = None,
        long: Optional[float] = None,
        units: Optional[str] = None,
        **kwargs
    ) -> List[dict]:
        """
        Returns the weather forecast for the upcoming hours/days.

        :param lat: Override the ``lat`` configuration value.
        :param long: Override the ``long`` configuration value.
        :param units: Override the ``units`` configuration value.
        :return: .. schema:: weather.openweathermap.WeatherSchema(many=True)
        """
        forecast = self._get_forecast(*args, lat=lat, long=long, units=units, **kwargs)

        if forecast:
            self._on_weather_forecast(forecast, always_publish=True)

        return forecast

    @action
    def status(self, *args, **kwargs) -> dict:
        """
        :return: .. schema:: weather.openweathermap.WeatherReportSchema
        """
        return self._status(*args, **kwargs)

    def _status(self, *args, **kwargs) -> dict:
        return dict(
            WeatherReportSchema().dump(
                {
                    'current': self.get_current_weather(*args, **kwargs).output,
                    'forecast': self.get_forecast(*args, **kwargs).output,
                }
            )
        )

    @abstractmethod
    def _get_current_weather(self, *args, **kwargs) -> dict:
        raise NotImplementedError("_get_current_weather not implemented")

    @abstractmethod
    def _get_forecast(self, *args, **kwargs) -> List[dict]:
        raise NotImplementedError("_get_forecast not implemented")

    def main(self):
        while not self.should_stop():
            try:
                self._status()
            except Exception as e:
                self.logger.exception(e)
            finally:
                self.wait_stop(self.poll_interval)
